import json
import os
import hashlib
import threading
import time
from datetime import datetime
from threading import Lock
from typing import (
    Dict,
    Optional,
    Tuple,
    List,
)

from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import (
    NotExpectedException,
    DisabledFunctionality,
    ErrorCreatingBackup,
    ErrorDeletingContent,
    RollbackError,
    NotInsideUpdateWindow,
    InvalidDownloadUpdate,
    NoUpdateFound,
    ErrorDownloadingLoaders,
    ErrorValidatingOpenedUsers,
)
from application.repository.api import BinaryUpdaterAPIRepository
from application.repository.pos import BinaryUpdaterPOSRepository
from application.service import (
    ApacheUpdaterService,
    StartStopUpdaterService,
    ChromeUpdaterService,
    LoaderUpdaterService,
    UpdaterService,
    CommonService,
)
from application.util import LoggerUtil, ValidationUtil, settings
from systools import sys_parsefilepath
from packaging import version

GENESIS_PATH = os.path.join("..", "..", "genesis")
SRC_PATH = os.path.join("..", "..", "src")

logger = LoggerUtil.get_logger_name(__name__)


class BinaryUpdaterInteractor(object):
    def __init__(
        self,
        configs,                     # type: Configurations
        working_on_update,           # type: Lock
        api_repository,              # type: BinaryUpdaterAPIRepository
        pos_repository,              # type: BinaryUpdaterPOSRepository
        updater_service,             # type: UpdaterService
        loader_updater_service,      # type: LoaderUpdaterService
        chrome_updater_service,      # type: ChromeUpdaterService
        start_stop_updater_service,  # type: StartStopUpdaterService
        apache_updater_service,      # type: ApacheUpdaterService
    ):
        # type: (...) -> None # noqa

        self.update_type = UpdateType.binaries
        self.update_name = self.update_type.name

        self.enabled = configs.updaters[self.update_name].enabled
        self.local_configs = configs.updaters[self.update_name]
        self.start_time = configs.updaters[self.update_name].apply_window.start_time
        self.end_time = configs.updaters[self.update_name].apply_window.end_time
        self.store_id = configs.store_id
        self.force_update_days = configs.updaters[self.update_name].force_update_days
        self.restrict_days_update = configs.updaters[self.update_name].restrict_days_update

        self.working_on_update = working_on_update
        self.api_repository = api_repository
        self.pos_repository = pos_repository
        self.updater_service = updater_service
        self.loader_updater_service = loader_updater_service

        data_path = sys_parsefilepath("{HVDATADIR}")
        work_dir = sys_parsefilepath(configs.backup_directory)
        self.htdocs_relative_path = "htdocs"
        self.htdocs_path = os.path.join(data_path, GENESIS_PATH, 'data', 'server', self.htdocs_relative_path)
        self.genesis_path = os.path.join(data_path, GENESIS_PATH)
        self.renamed_htdocs_path = self.htdocs_path + "_PRE_UPDATE"
        self.src_path = os.path.join(data_path, SRC_PATH)
        self.genesis_src_path = os.path.join(data_path, GENESIS_PATH, "src")
        self.renamed_src_path = self.genesis_src_path + "_PRE_UPDATE"
        self.genesis_bin_path = os.path.join(data_path, GENESIS_PATH, "bin")
        self.core_relative_linux_path = "linux-x86_64"
        self.core_linux_path = os.path.join(data_path, GENESIS_PATH, "bin", self.core_relative_linux_path)
        self.renamed_core_linux_path = self.core_linux_path + "_PRE_UPDATE"
        self.core_relative_windows_path = "windows-x86"
        self.core_windows_path = os.path.join(data_path, GENESIS_PATH, "bin", self.core_relative_windows_path)
        self.core_path = os.path.join(data_path, GENESIS_PATH, "bin")
        self.renamed_core_windows_path = self.core_windows_path + "_PRE_UPDATE"
        default_work_path = os.path.join(work_dir, self.update_type.name)
        self.downloads_path = os.path.join(default_work_path, "downloads")
        self.backups_path = os.path.join(default_work_path, "backups")
        self.work_path = os.path.join(default_work_path, "work_path")
        self.binary_src_zip = "binary_src.tar.gz"
        self.binary_core_zip = "binary_core.tar.gz"
        self.core_linux_backup = ''
        self.core_windows_backup = ''
        self.src_backup = ''
        self.htdocs_backup = ''
        self.python3_path = os.path.join(data_path, "..", "..", "python3")

        self.updater_service.create_dir(self.backups_path, False)
        self.updater_service.create_dir(self.downloads_path, False)
        self.apache_updater_service = apache_updater_service
        self.chrome_updater_service = chrome_updater_service
        self.start_stop_updater_service = start_stop_updater_service

    def update_binaries_start(self):
        result = self.pos_repository.get_last_binary_update_not_applied_controller_id()
        downloaded_date = result.get("downloaded_date", None)
        if downloaded_date is None:
            return

        diff_days = (datetime.now() - downloaded_date).days
        if self.force_update_days == 0 or (diff_days > self.force_update_days):
            t1 = threading.Thread(target=self.update_binaries, args=("updateNow", True))
            t1.start()

    def update_binaries(
        self,
        data,           # type: Optional[str]
        silent=False    # type: bool
    ):
        # type: (...) -> None

        settings.set_update_control(self.update_name, None)
        if not self.enabled:
            settings.set_update_control(self.update_name, "$UPDATE_DISABLED")
            raise DisabledFunctionality()

        rollback_necessary = False
        try:
            self.updater_service.create_dir(self.core_linux_path, False)
            self.updater_service.create_dir(self.core_windows_path, False)

            settings.set_update_control(self.update_name, "$UPDATING_BINARIES")
            settings.set_update_control("updating", not silent)
            backoffice_version = self.api_repository.check_last_update_binaries()

            core, core_version, src, src_version = self._has_update_to_apply(backoffice_version=backoffice_version)
            if self.pos_repository.has_already_applied_update() and silent and not self._can_update_in_current_day():
                settings.set_update_control(self.update_name, "$UPDATE_DISABLED_FOR_CURRENT_DAY")
                raise DisabledFunctionality()

            self.loader_updater_service.download_and_validate_loaders()

            logger.info("Waiting Updating Binaries")
            with self.working_on_update:
                core_new_version = backoffice_version["coreVersionNumber"]
                src_new_version = backoffice_version["srcVersionNumber"]

                file_src_path = os.path.join(self.downloads_path, src_new_version + self.binary_src_zip)
                file_core_path = os.path.join(self.downloads_path, core_new_version + self.binary_core_zip)
                result = self.pos_repository.get_last_binary_update_not_applied_controller_id()
                update_id = result.get("update_id", 0)
                downloaded = self._download_performed(
                    update_id=update_id,
                    backoffice_version=backoffice_version,
                    file_src_path=file_src_path,
                    file_core_path=file_core_path,
                )
                if not downloaded:
                    update_id = self._perform_download(
                        backoffice_version,
                        core_new_version,
                        file_core_path,
                        file_src_path,
                        src_new_version
                    )

                is_inside_allowed_window = ValidationUtil.is_inside_allowed_window(self.start_time, self.end_time)
                if (src_version and core_version) and (data != "updateNow" and not is_inside_allowed_window):
                    str_start_time = self.start_time.strftime("%H:%M:%S")
                    str_end_time = self.end_time.strftime("%H:%M:%S")
                    message = "Not inside allowed window. Start: {} / End: {}".format(str_start_time, str_end_time)
                    logger.info(message)
                    raise NotInsideUpdateWindow()

                settings.set_update_control("updating", True)
                self.updater_service.create_dir(self.work_path)

                self._create_backups(update_id)

                rollback_necessary = True
                if core:
                    self._apply_binary_core(file_core_path)

                if src:
                    self._apply_binary_src(file_src_path)

                rollback_necessary = False
                self.pos_repository.set_update_as_applied(
                    update_id=update_id,
                    backoffice_version=backoffice_version,
                )
            try:
                self.chrome_updater_service.update_chrome(
                    validate_version=True,
                    restart_hv=False
                )
            except (Exception,):
                logger.exception("Error updating chrome after Binaries update")

            try:
                self.loader_updater_service.update_loaders(
                    restart_hv=False,
                    retry_on_error=True
                )
            except (Exception,):
                logger.exception("Error updating loader after Binaries update")

            try:
                self.apache_updater_service.update_httpd_conf()
            except (Exception,):
                logger.exception("Error updating httpd.conf after Binaries update")

            try:
                self.start_stop_updater_service.update_start_stop()
            except (Exception,):
                logger.exception("Error updating start/stop files after Binaries update")

            self.updater_service.create_dir(self.downloads_path, True)
            logger.info("Starting Hypervisor to apply new Binary")
            self.pos_repository.restart_hv()
            time.sleep(5)

            logger.info("Binaries update fully concluded with success!")
        except NotInsideUpdateWindow:
            settings.set_update_control(self.update_name, "$NOT_INSIDE_UPDATE_WINDOW")
            logger.info("Sleeping until enter the update window")
            raise
        except ErrorValidatingOpenedUsers:
            settings.set_update_control(self.update_name, "$UPDATE_BINARIES_ERRORVALIDATINGOPENEDUSERS")
            logger.info("Error validating loaders. Necessary to close all the operators and the day")
            raise
        except NoUpdateFound:
            logger.info("No Binaries update found to apply")
            settings.set_update_control(self.update_name, "$SUCCESS_UPDATE")
            raise
        except (NotExpectedException, Exception) as ex:
            if rollback_necessary:
                with self.working_on_update:
                    self._rollback()

            if isinstance(ex, ErrorDownloadingLoaders):
                settings.set_update_control(self.update_name, ex.message)
            else:
                settings.set_update_control(self.update_name, "$UNKNOWN_ERROR")

            logger.exception("Not expected exception on binaries update")
            raise
        finally:
            self.updater_service.delete_directory(self.work_path)
            settings.set_update_control("updating", False)

    def _remove_non_applied_updates(self):
        update_id = 1
        while update_id != 0:
            result = self.pos_repository.get_last_binary_update_not_applied_controller_id()
            update_id = result.get("update_id", 0)
            if update_id != 0:
                self.pos_repository.remove_downloaded_update(update_id)

    def _has_update_to_apply(
        self,                   # type: BinaryUpdaterInteractor
        backoffice_version,     # type: Dict[str, str]
    ):
        # type: (...) -> Tuple[bool, str, bool, str]

        core_version, src_version = self._check_last_update_binaries()
        core_new_version = backoffice_version["coreVersionNumber"]
        core = self._check_if_update_is_different(core_version, core_new_version)
        src_new_version = backoffice_version["srcVersionNumber"]
        src = self._check_if_update_is_different(src_version, src_new_version)
        if not core and not src:
            logger.info("No update found")
            raise NoUpdateFound()

        return core, core_version, src, src_version

    def notify_update(
        self,       # type: BinaryUpdaterInteractor
    ):
        # type: (...) -> None

        update_data = self.pos_repository.get_last_binary_update_info()
        if not update_data.applied_date or update_data.notified_date:
            return

        server_update_identifier = self._get_server_update_identifier(update_data)
        if not server_update_identifier:
            return

        src_version = server_update_identifier.get("srcVersionNumber")
        src_pkg_info_file_version = self._get_pkg_info_file_version(
            pkg_info_file_path=os.path.join(self.src_path, "PKG-INFO")
        )
        if src_pkg_info_file_version is not None and src_version and src_pkg_info_file_version != src_version:
            return

        core_version = server_update_identifier.get("coreVersionNumber")
        core_pkg_info_file_version = self._get_pkg_info_file_version(
            pkg_info_file_path=os.path.join(self.core_path, "PKG-INFO")
        )
        if core_pkg_info_file_version is not None and core_version and core_pkg_info_file_version != core_version:
            return

        update_version_id = server_update_identifier.get("id")
        update_notified = self.api_repository.notify_update(
            version_id=str(update_version_id),
        )
        if update_notified:
            self.pos_repository.set_notified_date(
                update_id=str(update_data.update_id),
            )
            self._remove_non_applied_updates()

    def _get_server_update_identifier(self, update_data):
        server_update_identifier = json.loads(update_data.server_update_identifier or "{}")
        if server_update_identifier:
            return server_update_identifier

        backoffice_version = self.api_repository.check_last_update_binaries()
        backoffice_src_version = backoffice_version["srcVersionNumber"]
        database_src_version = update_data.update_name.split("|")[1].replace("SRC:", "")
        if backoffice_src_version != database_src_version:
            return {}

        return {
            "srcVersionNumber": backoffice_src_version,
            "id": backoffice_version["id"],
        }

    @staticmethod
    def _get_pkg_info_file_version(
        pkg_info_file_path,     # type: str
    ):
        # type: (...) -> Optional[str]

        if os.path.isfile(pkg_info_file_path):
            with open(pkg_info_file_path, 'rb') as f:
                lines = f.readlines()
                for line in lines:
                    if line.startswith("Version:"):
                        return line.split("Version:")[1].strip()

    def _create_backups(self, update_id):
        settings.set_update_control(self.update_name, "$CREATING_POS_BACKUP")
        self.core_windows_backup = self._create_backup(self.core_windows_path, "core_windows")
        self.core_linux_backup = self._create_backup(self.core_linux_path, "core_linux")
        self.src_backup = self._create_backup(self.genesis_src_path, "src")
        self.htdocs_backup = self._create_backup(self.htdocs_path, "htdocs")
        self.pos_repository.set_backup_date(update_id)

    def _perform_download(self, backoffice_version, core_new_version, file_core_path, file_src_path, src_new_version):
        logger.info("Updating Binaries")
        new_update_id = str(self.pos_repository.get_last_binary_update_id() + 1)
        update_name = "CORE:{}|SRC:{}".format(core_new_version, src_new_version)
        self.pos_repository.insert_new_binary_update(new_update_id, update_name)

        update_data = self.pos_repository.get_last_binary_update_info()
        logger.info("Downloading src")
        settings.set_update_control(self.update_name, "$DOWNLOADING_SRC_FROM_BACKOFFICE")
        self.api_repository.download_src(backoffice_version["srcVersionLink"], file_src_path)
        logger.info("SRC downloaded")

        logger.info("Downloading core")
        settings.set_update_control(self.update_name, "$DOWNLOADING_CORE_FROM_BACKOFFICE")
        self.api_repository.download_core(backoffice_version["coreVersionLink"], file_core_path)
        logger.info("CORE downloaded")

        hash_src = self._get_hash(file_src_path)
        hash_core = self._get_hash(file_core_path)
        src_version_hash = backoffice_version["srcVersionHash"]
        core_version_hash = backoffice_version["coreVersionHash"]

        if hash_src != src_version_hash or hash_core != core_version_hash:
            logger.info("POS srcVersionHash: {}".format(hash_src))
            logger.info("BackOffice srcVersionHash: {}".format(src_version_hash))
            logger.info("POS coreVersionHash: {}".format(hash_core))
            logger.info("BackOffice coreVersionHash: {}".format(core_version_hash))

            self.api_repository.delete_file(file_src_path)
            self.api_repository.delete_file(file_core_path)
            raise InvalidDownloadUpdate()

        file_src_hash_path = os.path.join(self.downloads_path, src_new_version + "-" + src_version_hash)
        self.api_repository.save_file(file_src_hash_path, "")
        file_core_hash_path = os.path.join(self.downloads_path, core_new_version + "-" + core_version_hash)
        self.api_repository.save_file(file_core_hash_path, "")

        self.pos_repository.set_downloaded_date(update_data.update_id)
        return update_data.update_id

    def _apply_binary_core(self, file_core_path):
        logger.info("Applying core")
        settings.set_update_control(self.update_name, "$APPLYING_CORE")

        un_gz_folder = os.path.join(self.work_path, "ungz")
        self.updater_service.create_dir(un_gz_folder)
        un_gz_folder = os.path.join(un_gz_folder, "core")
        self.updater_service.create_dir(un_gz_folder)

        self.updater_service.extract_tar_gz(file_core_path, un_gz_folder)
        core_update_folder_name = os.listdir(un_gz_folder)[0]

        for folder in ["apache", "bin", "python", "python3"]:
            for so in [self.core_relative_linux_path, self.core_relative_windows_path]:
                genesis_folder = os.path.join(self.genesis_path, folder)
                self.updater_service.create_dir(genesis_folder, False)
                genesis_so_path = os.path.join(genesis_folder, so)
                self.updater_service.create_dir(genesis_so_path)
                so_update_path = os.path.join(un_gz_folder, core_update_folder_name, "pos-core", so, folder)
                self.updater_service.copy_all_content(so_update_path, genesis_so_path)

        self._update_pkg_info_file(
            file_path=un_gz_folder,
            genesis_file=core_update_folder_name,
            destination_paths=[self.core_linux_path, self.core_windows_path],
        )

        logger.info("Core applied")

    def _apply_binary_src(self, file_src_path):
        logger.info("Applying src")
        settings.set_update_control(self.update_name, "$APPLYING_SRC")

        if os.path.exists(self.genesis_src_path):
            self.updater_service.rename_folder(self.genesis_src_path, self.renamed_src_path)

        if os.path.exists(self.htdocs_path):
            self.updater_service.rename_folder(self.htdocs_path, self.renamed_htdocs_path)

        self.updater_service.create_dir(self.htdocs_path)

        ungz_folder = os.path.join(self.work_path, "ungz")
        self.updater_service.create_dir(ungz_folder)
        ungz_folder = os.path.join(ungz_folder, "src")
        self.updater_service.create_dir(ungz_folder)

        self.updater_service.extract_tar_gz(file_src_path, ungz_folder)
        src_update_folder_name = os.listdir(ungz_folder)[0]

        temp_path = os.path.join(ungz_folder, src_update_folder_name, "pos-src", "_comps")
        self.updater_service.move_directory(temp_path, self.genesis_src_path)

        gui_zip_folder = os.path.join(ungz_folder, src_update_folder_name, "pos-src", "_guizip")
        self.updater_service.extract_zip(os.path.join(gui_zip_folder, "sui.zip"), os.path.join(self.htdocs_path, "sui"))
        self.updater_service.extract_zip(os.path.join(gui_zip_folder, "kui.zip"), os.path.join(self.htdocs_path, "kui"))
        self.updater_service.extract_zip(os.path.join(gui_zip_folder, "nti.zip"), os.path.join(self.htdocs_path, "nti"))

        temp_path = os.path.join(ungz_folder, src_update_folder_name, "pos-src", "_pypkg")
        self.updater_service.copy_all_files(temp_path, self.core_windows_path)
        self.updater_service.copy_all_files(temp_path, self.core_linux_path)

        self._update_pkg_info_file(
            file_path=ungz_folder,
            genesis_file=src_update_folder_name,
            destination_paths=[self.genesis_src_path],
        )

        self.updater_service.delete_directory(self.renamed_src_path)
        self.updater_service.delete_directory(self.renamed_htdocs_path)

        logger.info("Src applied")

    def _update_pkg_info_file(
        self,                   # type: BinaryUpdaterInteractor
        file_path,              # type: str
        genesis_file,           # type: str
        destination_paths,      # type: List[str]
    ):
        # type: (...) -> None

        try:
            pkg_info_file = os.path.join(file_path, genesis_file, "PKG-INFO")
            if not os.path.isfile(pkg_info_file):
                logger.info("No PKG-INFO file found")
                return

            for path in destination_paths:
                self.updater_service.copy_file(
                    original_file_path=pkg_info_file,
                    destination_file_path=path,
                )
        except (Exception,):
            logger.exception("Error updating PKG-INFO file")

    def _check_last_update_binaries(self):
        last_update = self.pos_repository.get_last_binary_update_name()
        core_version = last_update.split('|')[0].split(':')[1] if last_update else ""
        src_version = last_update.split('|')[1].split(':')[1] if last_update else ""

        return core_version, src_version

    @staticmethod
    def _check_if_update_is_different(last_update, new_update):
        return version.parse(new_update) > version.parse(last_update)

    @staticmethod
    def _get_hash(file_path):
        # BUF_SIZE is totally arbitrary, change for your app!
        buf_size = 65536  # lets read stuff in 64kb chunks!

        sha256 = hashlib.sha256()

        with open(file_path, 'rb') as f:
            while True:
                data = f.read(buf_size)
                if not data:
                    break
                sha256.update(data)

        return sha256.hexdigest()

    def _create_backup(self, path_to_backup, update_type):
        # type: (str, str) -> str

        try:
            logger.info("Creating {} backup".format(update_type))

            backup_zip_name = "binaries_{}_{}".format(update_type, datetime.now().strftime("%Y%m%d_%H%M%S"))
            backup_package = os.path.join(self.backups_path, backup_zip_name)
            backup_package_zip = os.path.join(self.backups_path, backup_zip_name) + ".zip"

            zip_path = os.path.join(self.work_path, update_type)
            if os.path.exists(path_to_backup):
                self.updater_service.copy_directory(path_to_backup, zip_path)
                self.updater_service.save_directory_zip(zip_path, backup_package)
                self.updater_service.test_zip_file(backup_package_zip)
                self._clean_backups()

            logger.info("Backup successfully created for {}!".format(update_type))
            return backup_package_zip
        except (Exception,):
            logger.exception("Error creating {} backup".format(update_type))
            raise ErrorCreatingBackup()

    def _clean_backups(self):
        # type: () -> None

        list_of_files = os.listdir(self.backups_path)

        while len(list_of_files) > 16:
            full_path = [os.path.join(self.backups_path, x) for x in list_of_files]
            oldest_file = min(full_path, key=os.path.getctime)
            if os.path.isfile(oldest_file):
                self.updater_service.delete_file(oldest_file)
            else:
                self.updater_service.delete_directory(oldest_file)

            list_of_files = os.listdir(self.backups_path)

    def _delete_content(self, path):
        # type: (str) -> None

        try:
            logger.info("Deleting path {}".format(path))

            self.updater_service.delete_directory(path)

            logger.info("Path deleted")
        except (Exception,):
            logger.exception("Error deleting path")
            raise ErrorDeletingContent()

    def _rollback(self):
        # type: () -> None

        logger.info("Starting rollback")

        retries = 0
        max_retries = 2

        while True:
            try:
                if self.core_windows_backup != '':
                    self._delete_content(self.core_windows_path)
                    self._delete_content(self.renamed_core_windows_path)
                    self.updater_service.extract_zip(self.core_windows_backup, self.core_windows_path)
                if self.core_linux_backup != '':
                    self._delete_content(self.core_linux_path)
                    self._delete_content(self.renamed_core_linux_path)
                    self.updater_service.extract_zip(self.core_linux_backup, self.core_linux_path)
                if self.src_backup != '' and os.path.exists(self.genesis_src_path):
                    self._delete_content(self.genesis_src_path)
                    self._delete_content(self.renamed_src_path)
                    self.updater_service.extract_zip(self.src_backup, self.genesis_src_path)
                if self.htdocs_backup != '':
                    self._delete_content(self.htdocs_path)
                    self._delete_content(self.renamed_htdocs_path)
                    self.updater_service.extract_zip(self.htdocs_backup, self.htdocs_path)

            except (Exception,):
                logger.exception("Error while executing rollback")
                if retries == max_retries:
                    raise RollbackError()

                retries += 1
                continue

            logger.info("Rollback finished")
            break

    def _rename_path(self, path, renamed_path):
        # type: (str, str) -> None

        try:
            logger.info("Renaming folder from {} to {}".format(path, renamed_path))

            self.updater_service.rename_folder(path, renamed_path)

            logger.info("Data folder renamed!")
        except (Exception,):
            logger.exception("Error renaming data folder")
            raise ErrorDeletingContent()

    def _download_performed(self, update_id, backoffice_version, file_src_path, file_core_path):
        if update_id == 0:
            return False

        src_new_version = backoffice_version["srcVersionNumber"]
        src_version_hash = backoffice_version["srcVersionHash"]
        file_src_hash_path = os.path.join(self.downloads_path, src_new_version + "-" + src_version_hash)
        src_hash_ok = os.path.exists(file_src_hash_path)

        core_new_version = backoffice_version["coreVersionNumber"]
        core_version_hash = backoffice_version["coreVersionHash"]
        file_core_hash_path = os.path.join(self.downloads_path, core_new_version + "-" + core_version_hash)
        core_hash_ok = os.path.exists(file_core_hash_path)

        has_src = os.path.exists(file_src_path) and src_hash_ok
        has_core = os.path.exists(file_core_path) and core_hash_ok
        if has_src and has_core:
            return True

        self.api_repository.delete_file(file_src_hash_path)
        self.api_repository.delete_file(file_core_hash_path)
        self.api_repository.delete_file(file_src_path)
        self.api_repository.delete_file(file_core_path)
        self.pos_repository.remove_downloaded_update(update_id)

        return False

    def has_binary_update(
        self
    ):
        # type: (...) -> Dict

        result = self.pos_repository.get_last_binary_update_not_applied_controller_id()
        update_id = result.get("update_id", 0)

        if update_id == 0 or not self._can_update_in_current_day():
            return {"update": "NO_PENDING_UPDATE"}
        else:
            if self.force_update_days == 0:
                return {"update": "UPDATE_NOW"}

            downloaded_date = result.get("downloaded_date", None)
            if downloaded_date is not None and (datetime.now() - downloaded_date).days >= self.force_update_days:
                return {"update": "UPDATE_NOW"}

            last_update_name = self.pos_repository.get_last_binary_update_name()
            current_update_name = result.get("update_name", None)
            return {
                "update": "HAS_PENDING_UPDATE",
                "lastVersion": last_update_name,
                "currentVersion": current_update_name
            }

    def _can_update_in_current_day(
        self
    ):
        # type: (...)-> bool

        weekday = datetime.today().weekday()
        return weekday not in self.restrict_days_update

    def has_python3_path(
        self
    ):
        # type: () -> bool

        if CommonService().check_is_dev():
            return True

        return os.path.exists(self.python3_path)

    def reapply_last_binary_update(
        self,  # type: BinaryUpdaterInteractor
    ):
        # type: (...) -> None

        try:
            self.pos_repository.invalidate_last_binary_update()
            self.update_binaries(
                data="updateNow",
            )
        except (Exception,):
            logger.exception("Error reapplying last binary")



from typing import List

from devicemanager.interactor.dto import DeviceStatusDto


class UpdateDevicesStatusRequest(object):
    def __init__(
        self,
        devices,    # type: List[DeviceStatusDto]
        operator,   # type: str
    ):
        # type: (...) -> None

        self.devices = devices
        self.operator = operator





class OperatorSession(object):

    def __init__(
        self,           # type: OperatorSession
        pos_id,         # type: str
        session_id,     # type: str
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.session_id = session_id



class LoyaltyConfigurations(object):

    def __init__(
        self,
        print_loyalty_coupon_after_fiscalization,   # type: bool
        print_loyalty_coupon_delivery_orders,       # type: bool
        text_to_print_in_loyalty_coupon,            # type: unicode
        store_name,                                 # type: unicode
    ):
        # type: (...) -> None

        self.print_loyalty_coupon_after_fiscalization = print_loyalty_coupon_after_fiscalization
        self.print_loyalty_coupon_delivery_orders = print_loyalty_coupon_delivery_orders
        self.text_to_print_in_loyalty_coupon = text_to_print_in_loyalty_coupon
        self.store_name = store_name



import json
from typing import List

from messagebus import (
    AckMessage,
    InternalErrorMessage,
)
from messageprocessorutil import UuidMessageProcessor
from model import EventPublishDetail
from model.mapper import EventPublishDetailMapper
from model.repository import EventRepository


class GetPendingProcessor(UuidMessageProcessor):
    def __init__(self, event_repository: EventRepository):
        super().__init__()

        self.event_repository = event_repository

    def parse_data(self, **kwargs):
        return None

    def call_business(self, _):
        return self.event_repository.get_pending()

    def format_response(
        self,
        message_bus,
        message,
        event,
        input_obj,
        pending_events: List[EventPublishDetail],
    ):
        ret = []
        for pending in pending_events:
            ret.append(EventPublishDetailMapper.map(publish_detail=pending).to_dict())

        message_bus.reply_message(
            message=message,
            reply_message=AckMessage(
                data=json.dumps(ret).encode("utf-8")
            )
        )

    def format_exception(
        self,
        message_bus,
        message,
        event,
        input_obj,
        exception
    ):
        message_bus.reply_message(
            message=message,
            reply_message=InternalErrorMessage(ex=exception)
        )

    def format_parse_exception(
        self,
        message_bus,
        message,
        event,
        data,
        exception
    ):
        return self.format_exception(
            message_bus=message_bus,
            message=message,
            event=event,
            input_obj=data,
            exception=exception,
        )



from pyasn1.type import base, tag

class EndOfOctets(base.AbstractSimpleAsn1Item):
    defaultValue = 0
    tagSet = tag.initTagSet(
        tag.Tag(tag.tagClassUniversal, tag.tagFormatSimple, 0x00)
        )
endOfOctets = EndOfOctets()



from __future__ import annotations

from typing import Any
from typing import ClassVar
from typing import Dict
from typing import Generic
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from sqlalchemy.sql.schema import Constraint
from sqlalchemy.sql.schema import ForeignKeyConstraint
from sqlalchemy.sql.schema import Index
from sqlalchemy.sql.schema import UniqueConstraint
from typing_extensions import TypeGuard

from alembic.ddl.base import _fk_spec
from .. import util
from ..util import sqla_compat

if TYPE_CHECKING:
    from typing import Literal

    from alembic.autogenerate.api import AutogenContext
    from alembic.ddl.impl import DefaultImpl

CompareConstraintType = Union[Constraint, Index]

_C = TypeVar("_C", bound=CompareConstraintType)

_clsreg: Dict[str, Type[_constraint_sig]] = {}


class ComparisonResult(NamedTuple):
    status: Literal["equal", "different", "skip"]
    message: str

    @property
    def is_equal(self) -> bool:
        return self.status == "equal"

    @property
    def is_different(self) -> bool:
        return self.status == "different"

    @property
    def is_skip(self) -> bool:
        return self.status == "skip"

    @classmethod
    def Equal(cls) -> ComparisonResult:
        """the constraints are equal."""
        return cls("equal", "The two constraints are equal")

    @classmethod
    def Different(cls, reason: Union[str, Sequence[str]]) -> ComparisonResult:
        """the constraints are different for the provided reason(s)."""
        return cls("different", ", ".join(util.to_list(reason)))

    @classmethod
    def Skip(cls, reason: Union[str, Sequence[str]]) -> ComparisonResult:
        """the constraint cannot be compared for the provided reason(s).

        The message is logged, but the constraints will be otherwise
        considered equal, meaning that no migration command will be
        generated.
        """
        return cls("skip", ", ".join(util.to_list(reason)))


class _constraint_sig(Generic[_C]):
    const: _C

    _sig: Tuple[Any, ...]
    name: Optional[sqla_compat._ConstraintNameDefined]

    impl: DefaultImpl

    _is_index: ClassVar[bool] = False
    _is_fk: ClassVar[bool] = False
    _is_uq: ClassVar[bool] = False

    _is_metadata: bool

    def __init_subclass__(cls) -> None:
        cls._register()

    @classmethod
    def _register(cls):
        raise NotImplementedError()

    def __init__(
        self, is_metadata: bool, impl: DefaultImpl, const: _C
    ) -> None:
        raise NotImplementedError()

    def compare_to_reflected(
        self, other: _constraint_sig[Any]
    ) -> ComparisonResult:
        assert self.impl is other.impl
        assert self._is_metadata
        assert not other._is_metadata

        return self._compare_to_reflected(other)

    def _compare_to_reflected(
        self, other: _constraint_sig[_C]
    ) -> ComparisonResult:
        raise NotImplementedError()

    @classmethod
    def from_constraint(
        cls, is_metadata: bool, impl: DefaultImpl, constraint: _C
    ) -> _constraint_sig[_C]:
        # these could be cached by constraint/impl, however, if the
        # constraint is modified in place, then the sig is wrong.  the mysql
        # impl currently does this, and if we fixed that we can't be sure
        # someone else might do it too, so play it safe.
        sig = _clsreg[constraint.__visit_name__](is_metadata, impl, constraint)
        return sig

    def md_name_to_sql_name(self, context: AutogenContext) -> Optional[str]:
        return sqla_compat._get_constraint_final_name(
            self.const, context.dialect
        )

    @util.memoized_property
    def is_named(self):
        return sqla_compat._constraint_is_named(self.const, self.impl.dialect)

    @util.memoized_property
    def unnamed(self) -> Tuple[Any, ...]:
        return self._sig

    @util.memoized_property
    def unnamed_no_options(self) -> Tuple[Any, ...]:
        raise NotImplementedError()

    @util.memoized_property
    def _full_sig(self) -> Tuple[Any, ...]:
        return (self.name,) + self.unnamed

    def __eq__(self, other) -> bool:
        return self._full_sig == other._full_sig

    def __ne__(self, other) -> bool:
        return self._full_sig != other._full_sig

    def __hash__(self) -> int:
        return hash(self._full_sig)


class _uq_constraint_sig(_constraint_sig[UniqueConstraint]):
    _is_uq = True

    @classmethod
    def _register(cls) -> None:
        _clsreg["unique_constraint"] = cls

    is_unique = True

    def __init__(
        self,
        is_metadata: bool,
        impl: DefaultImpl,
        const: UniqueConstraint,
    ) -> None:
        self.impl = impl
        self.const = const
        self.name = sqla_compat.constraint_name_or_none(const.name)
        self._sig = tuple(sorted([col.name for col in const.columns]))
        self._is_metadata = is_metadata

    @property
    def column_names(self) -> Tuple[str, ...]:
        return tuple([col.name for col in self.const.columns])

    def _compare_to_reflected(
        self, other: _constraint_sig[_C]
    ) -> ComparisonResult:
        assert self._is_metadata
        metadata_obj = self
        conn_obj = other

        assert is_uq_sig(conn_obj)
        return self.impl.compare_unique_constraint(
            metadata_obj.const, conn_obj.const
        )


class _ix_constraint_sig(_constraint_sig[Index]):
    _is_index = True

    name: sqla_compat._ConstraintName

    @classmethod
    def _register(cls) -> None:
        _clsreg["index"] = cls

    def __init__(
        self, is_metadata: bool, impl: DefaultImpl, const: Index
    ) -> None:
        self.impl = impl
        self.const = const
        self.name = const.name
        self.is_unique = bool(const.unique)
        self._is_metadata = is_metadata

    def _compare_to_reflected(
        self, other: _constraint_sig[_C]
    ) -> ComparisonResult:
        assert self._is_metadata
        metadata_obj = self
        conn_obj = other

        assert is_index_sig(conn_obj)
        return self.impl.compare_indexes(metadata_obj.const, conn_obj.const)

    @util.memoized_property
    def has_expressions(self):
        return sqla_compat.is_expression_index(self.const)

    @util.memoized_property
    def column_names(self) -> Tuple[str, ...]:
        return tuple([col.name for col in self.const.columns])

    @util.memoized_property
    def column_names_optional(self) -> Tuple[Optional[str], ...]:
        return tuple(
            [getattr(col, "name", None) for col in self.const.expressions]
        )

    @util.memoized_property
    def is_named(self):
        return True

    @util.memoized_property
    def unnamed(self):
        return (self.is_unique,) + self.column_names_optional


class _fk_constraint_sig(_constraint_sig[ForeignKeyConstraint]):
    _is_fk = True

    @classmethod
    def _register(cls) -> None:
        _clsreg["foreign_key_constraint"] = cls

    def __init__(
        self,
        is_metadata: bool,
        impl: DefaultImpl,
        const: ForeignKeyConstraint,
    ) -> None:
        self._is_metadata = is_metadata

        self.impl = impl
        self.const = const

        self.name = sqla_compat.constraint_name_or_none(const.name)

        (
            self.source_schema,
            self.source_table,
            self.source_columns,
            self.target_schema,
            self.target_table,
            self.target_columns,
            onupdate,
            ondelete,
            deferrable,
            initially,
        ) = _fk_spec(const)

        self._sig: Tuple[Any, ...] = (
            self.source_schema,
            self.source_table,
            tuple(self.source_columns),
            self.target_schema,
            self.target_table,
            tuple(self.target_columns),
        ) + (
            (None if onupdate.lower() == "no action" else onupdate.lower())
            if onupdate
            else None,
            (None if ondelete.lower() == "no action" else ondelete.lower())
            if ondelete
            else None,
            # convert initially + deferrable into one three-state value
            "initially_deferrable"
            if initially and initially.lower() == "deferred"
            else "deferrable"
            if deferrable
            else "not deferrable",
        )

    @util.memoized_property
    def unnamed_no_options(self):
        return (
            self.source_schema,
            self.source_table,
            tuple(self.source_columns),
            self.target_schema,
            self.target_table,
            tuple(self.target_columns),
        )


def is_index_sig(sig: _constraint_sig) -> TypeGuard[_ix_constraint_sig]:
    return sig._is_index


def is_uq_sig(sig: _constraint_sig) -> TypeGuard[_uq_constraint_sig]:
    return sig._is_uq


def is_fk_sig(sig: _constraint_sig) -> TypeGuard[_fk_constraint_sig]:
    return sig._is_fk



import time
import logging
from pyasn1.compat.octets import octs2ints
from pyasn1 import error
from pyasn1 import __version__

flagNone     = 0x0000
flagEncoder  = 0x0001
flagDecoder  = 0x0002
flagAll      = 0xffff

flagMap = {
    'encoder': flagEncoder,
    'decoder': flagDecoder,
    'all': flagAll
    }

class Printer:
    def __init__(self, logger=None, handler=None, formatter=None):
        if logger is None:
            logger = logging.getLogger('pyasn1')
        logger.setLevel(logging.DEBUG)
        if handler is None:
            handler = logging.StreamHandler()
        if formatter is None:
            formatter = logging.Formatter('%(asctime)s %(name)s: %(message)s')
        handler.setFormatter(formatter)
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        self.__logger = logger

    def __call__(self, msg): self.__logger.debug(msg)
    def __str__(self): return '<python built-in logging>'

if hasattr(logging, 'NullHandler'):
    NullHandler = logging.NullHandler
else:
    # Python 2.6 and older
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

class Debug:
    defaultPrinter = None
    def __init__(self, *flags, **options):
        self._flags = flagNone
        if options.get('printer') is not None:
            self._printer = options.get('printer')
        elif self.defaultPrinter is not None:
            self._printer = self.defaultPrinter
        if 'loggerName' in options: 
            # route our logs to parent logger
            self._printer = Printer(
                logger=logging.getLogger(options['loggerName']),
                handler=NullHandler()
            )
        else:
            self._printer = Printer()
        self('running pyasn1 version %s' % __version__)
        for f in flags:
            inverse = f and f[0] in ('!', '~')
            if inverse:
                f = f[1:]
            try:
                if inverse:
                    self._flags &= ~flagMap[f]
                else:
                    self._flags |= flagMap[f]
            except KeyError:
                raise error.PyAsn1Error('bad debug flag %s' % f)
  
            self('debug category \'%s\' %s' % (f, inverse and 'disabled' or 'enabled'))

    def __str__(self):
        return 'logger %s, flags %x' % (self._printer, self._flags)
    
    def __call__(self, msg):
        self._printer(msg)

    def __and__(self, flag):
        return self._flags & flag

    def __rand__(self, flag):
        return flag & self._flags

logger = 0

def setLogger(l):
    global logger
    logger = l

def hexdump(octets):
    return ' '.join(
            [ '%s%.2X' % (n%16 == 0 and ('\n%.5d: ' % n) or '', x) 
              for n,x in zip(range(len(octets)), octs2ints(octets)) ]
        )

class Scope:
    def __init__(self):
        self._list = []

    def __str__(self): return '.'.join(self._list)

    def push(self, token):
        self._list.append(token)

    def pop(self):
        return self._list.pop()

scope = Scope()



from abc import ABCMeta, abstractmethod
from logging import Logger

from production.model import ProductionOrder, ProdStates
from typing import Optional, Union, List, Callable

from .._ProductionBox import ProductionBox
from ...manager import StatisticUpdater


class StatisticBox(ProductionBox):
    __metaclass__ = ABCMeta

    def __init__(self, name, sons, statistic_updater, logger, get_config):
        # type: (str, Optional[Union[List[ProductionBox]], ProductionBox], StatisticUpdater, Logger, Callable[[str], str]) -> None # noqa

        super(StatisticBox, self).__init__(name, sons, logger)
        self.statistic_updater = statistic_updater
        self.statistic_name = get_config("StatisticName")

    def order_modified(self, order):
        # type: (ProductionOrder) -> None

        if order.prod_state != ProdStates.INVALID:
            new_statistic = self.update_statistics(order)
            if new_statistic is not None:
                self.statistic_updater.update_statistics(order.order_id, {self.statistic_name: new_statistic})
                self.debug('statistic "{}" updated to "{}"'.format(self.statistic_name, new_statistic))
        self.send_to_children(order)

    @abstractmethod
    def update_statistics(self, order):
        # type: (ProductionOrder) -> Optional[float]
        raise NotImplementedError()



from typing import List


class GoalsCache(object):
    def __init__(
        self,
        store_goals,    # type: object
        item_goals,     # type: List
    ):
        # type: (...) -> None

        self.store_goals = store_goals
        self.item_goals = item_goals



from abc import ABCMeta, abstractmethod
from datetime import datetime  # noqa
from typing import Optional  # noqa


class ShowKeyboard(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def show(self, pos_id, message, title='$INPUT_DATA',
             buttons='$OK|$CANCEL', icon='info', timeout=60000, mask='', is_password=False,
             def_value='', min_value='', maxvalue='', num_pad=False, no_pad=False, no_input=False,
             asynch=False):
        # type: (str, str, Optional[str], Optional[str], Optional[str], Optional[str], Optional[int], Optional[str], Optional[bool], Optional[str], Optional[str], Optional[str], Optional[bool], Optional[bool], Optional[bool], Optional[bool]) -> str
        raise NotImplementedError()

    @abstractmethod
    def show_date(self, pos_id, message, title="", show_message_wrong_input_date=False):
        # type: (str, str, Optional[str], Optional[bool]) -> datetime
        raise NotImplementedError()



import logging

from cfgtools import (
    read,
    Group,
)
from typing import (
    Dict,
    List,
    Tuple,
)

from devicemanager.model import (
    Configs,
    DeviceConfig,
)

logger = logging.getLogger(__name__)


class Configurator(object):

    @staticmethod
    def read_loader_configs(
        loader_path,  # type: str
    ):
        # type: (...) -> Configs

        reader = read(loader_path)
        main_group = reader.find_group("DeviceManager")

        device_configs = Configurator._parse_device_configs(
            main_group=main_group,
            config_names=["POS", "KDS"]
        )
        brand_config = Configurator._get_brands_configs(
            main_group=main_group,
        )

        return Configs(
            component_name=reader.find_value("Identification.ComponentName"),
            required_services="",
            persistence_name=reader.find_value("DeviceManager.PersistenceName").encode("utf-8"),
            sector_config=Configurator._get_sectors(main_group=main_group),
            device_configs=device_configs,
            brand_config=brand_config,
        )

    @staticmethod
    def _get_sectors(
        main_group,  # type: Group
    ):
        # type: (...) -> Dict[str, Dict[str, Tuple[str, Dict[int, int]]]]

        sectors_config = {}
        sectors = main_group.find_group("Sectors")
        for sector in sectors.groups:
            sector_name = sector.name.upper()
            if sector_name not in sectors_config:
                sectors_config[sector_name] = {}

            excluded_pos = {}
            for key in sector.keys:
                key_name = key.name.upper()
                if key_name == "EXCLUDEDPOS":
                    for pos_number in key.values:
                        excluded_pos[int(pos_number)] = int(pos_number)
                else:
                    device_type = key_name
                    device_name = key.value()
                    if device_type in sectors_config[sector_name]:
                        logger.info("Device manager sector loader key duplicated - {}{}".format(sector.name, key.name))
                    sectors_config[sector_name][device_type] = (device_name, excluded_pos)

            if not sectors_config[sector_name]:
                del sectors_config[sector_name]

        return sectors_config

    @staticmethod
    def _parse_device_configs(
        main_group,    # type: Group
        config_names,  # type: List[str]
    ):
        # type: (...) -> List[DeviceConfig]

        device_configs = []
        for config_name in config_names:
            device_groups = main_group.find_group(config_name)
            if not device_groups:
                continue

            for device_group in device_groups.groups:
                group_name = device_group.name.upper()
                printer_group = device_group.find_group("printer")
                if not printer_group:
                    continue

                default_device = printer_group.find_value("Default")
                fallback_devices = printer_group.find_values("Fallback") or []

                device_configs.append(
                    DeviceConfig(
                        requester_type=group_name[:3],
                        requester_id=int(group_name[3:]),
                        default_device=default_device,
                        fallback_devices=fallback_devices,
                    )
                )
        return device_configs

    @staticmethod
    def _get_brands_configs(
        main_group,    # type: Group
    ):
        # type: (...) -> Dict[str, Dict[str, str]]

        brand_config = {}
        brands = main_group.find_group("Brands")
        if not brands:
            return brand_config

        for brand in brands.groups:
            if not brand.keys:
                continue

            brand_name = brand.name.upper()
            for brand_key in brand.keys:
                brand_config.update(
                    {
                        brand_name: {
                            brand_key.name.upper(): brand_key.value()
                        }
                    }

                )

        return brand_config



from report import Formatter
from report.command import BoldCommand, NewLineCommand, AlignCommand, RepeatCommand


class HtmlFormatter(Formatter):
    def __init__(self):
        self.bold_start_index = -1
        self.bold_end_index = -1
        self.is_bold = False
        self.current_alignment = AlignCommand.Alignment.left

        self.current_line = ""
        self.body = ""

    def format(self, report):
        self.handle_new_line(report)

        for part in report.get_parts():
            if part.commands is not None:
                for command in part.commands:
                    if isinstance(command, BoldCommand):
                        self.handle_bold(command)
                    if isinstance(command, NewLineCommand):
                        self.handle_new_line(report)
                    if isinstance(command, AlignCommand):
                        self.current_alignment = command.alignment
                    if isinstance(command, RepeatCommand):
                        self.current_line += command.text * command.count

            if part.text is not None:
                if (part.commands is None or len(part.commands) == 0) and part.text == "\n":
                    self.handle_new_line(report)
                else:
                    self.current_line += part.text

        if self.current_line != "":
            self.handle_new_line(report)

        return self.body

    def handle_bold(self, command):
        if command.activate and not self.is_bold:
            self.bold_start_index = len(self.current_line)
            self.is_bold = True
        if not command.activate and self.is_bold:
            self.bold_end_index = len(self.current_line)
            self.is_bold = False

    def handle_new_line(self, report):
        formatted_text = self.create_formatted_text(self.current_line, report)
        self.add_line_to_body(formatted_text)
        self.current_line = ""

    def add_line_to_body(self, formatted_text):
        self.body += "<span style='font-family: monospace;font-size: 1.3em'>" + formatted_text + "</span>"
        self.body += "<br/>"

    def create_formatted_text(self, line, report):
        formatted_text = line
        if len(line) > report.get_width():
            formatted_text = line[0: report.get_width()]
        elif len(line) < report.get_width():
            if self.current_alignment == AlignCommand.Alignment.left:
                formatted_text += (" " * (report.get_width() - len(line)))
            elif self.current_alignment == AlignCommand.Alignment.right:
                formatted_text = (" " * (report.get_width() - len(line))) + formatted_text
            else:
                pre_space = int((report.get_width() - len(line)) / 2)
                post_space = report.get_width() - len(line) - pre_space
                formatted_text = (" " * pre_space) + formatted_text + (" " * post_space)
        before_bold = formatted_text
        bold_text = ""
        after_bold = ""
        if self.bold_start_index >= 0:
            if self.bold_end_index >= 0:
                before_bold = formatted_text[0: self.bold_start_index]
                bold_text = formatted_text[self.bold_start_index: self.bold_end_index]
                after_bold = formatted_text[self.bold_end_index:]

                self.bold_start_index = -1
                self.bold_end_index = -1

            else:
                before_bold = formatted_text[0: self.bold_start_index]
                bold_text = formatted_text[self.bold_start_index:]
                after_bold = ""

                self.bold_start_index = 0
        formatted_text = before_bold.replace(" ", "&nbsp;")
        if bold_text != "":
            formatted_text += "<span style='font-weight: bold;font-family: monospace;font-size: 1.3em'>" + bold_text.replace(" ", "&nbsp;") + "</span>"
        formatted_text += after_bold.replace(" ", "&nbsp;")

        return formatted_text



import math
import re

from abc import ABCMeta
from datetime import (
    datetime,
    timedelta,
)
from typing import List, Any, Callable, Tuple

from ._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import ProductionOrder, Item

DATE_FORMAT = "%Y-%m-%dT%H:%M:%S"


class OrderSplitterBox(OrderChangerProductionBox):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        name,
        sons,
        order_changer,
        publish_scheduler,
        split_type,
        split_value,
        logger=None
    ):
        super(OrderSplitterBox, self).__init__(name, sons, logger)
        self.order_changer = order_changer
        self.publish_scheduler = publish_scheduler
        self.split_type = split_type
        self.split_value = split_value

    def change_order(self, order):
        # type: (ProductionOrder) -> List[ProductionOrder]

        if self.split_type is None:
            return [order]
        else:
            if self.split_type.upper() == 'PREP_TIME':
                orders = self.split_prep_time(order)
            elif self.split_type.upper() == 'SALE_LINES':
                if self.split_value is None:
                    return [order]

                orders = self.split_sale_line(order, float(self.split_value))
            elif self.split_type.upper() in 'FIRE_HOLD_PRIORITY':
                orders = self.split_fire_hold_priority(order)
                self._schedule_hold_orders(orders)
                self._remove_split_tags(orders)
                self._add_fire_hold_tag(orders)
            else:
                orders = [order]

        return orders

    @staticmethod
    def split_prep_time(order):
        # type: (ProductionOrder) -> List[ProductionOrder]

        orders = []
        split_index = 0

        for item in order.items:
            for temp_order in orders:
                if len(temp_order.items) > 0 and item.prep_time == temp_order.items[0].prep_time:
                    temp_order.items.append(item)
                    break
            else:
                new_order = order.clone([item], split_index)
                orders.append(new_order)
                split_index += 1

        return orders

    def split_sale_line(self, order, limit):
        # type: (ProductionOrder, float) -> List[ProductionOrder]

        temp_items = [x for x in order.items if not self._has_fire_hold_or_priority(x)]
        if len(temp_items) < limit:
            return [order]

        orders = []
        split_index = 0
        box_limit = math.ceil(len(temp_items) / limit)
        items_per_box = math.ceil(len(temp_items) / box_limit)
        for item in temp_items:
            if len(orders) <= split_index:
                new_order = order.clone([], split_index)
                orders.append(new_order)

            orders[split_index].items.append(item)
            if len(orders[split_index].items) >= items_per_box:
                split_index += 1

        orders[0].items.extend([x for x in order.items if self._has_fire_hold_or_priority(x)])
        return orders

    def split_fire_hold_priority(self, order):
        # type: (ProductionOrder) -> List[ProductionOrder]

        orders = []
        main_order_items = []
        split_index = 1
        for item in order.items:
            if not self._has_fire_hold_or_priority(item):
                main_order_items.append(item)
                continue

            hold_item, split_index = self._handle_hold_item(
                item,
                order,
                orders,
                split_index
            )

            fire_item, split_index = self._handle_fire(
                item,
                order,
                orders,
                split_index
            )

            priority_item, split_index = self._handle_priority(
                item,
                order,
                orders,
                split_index
            )

            if not hold_item:
                if fire_item or priority_item:
                    property_list = [self._fire_in_item, self._priority_in_item]
                    clear_item = self._get_item_without_properties(item.clone(), property_list)
                    if clear_item:
                        main_order_items.append(clear_item)
            else:
                if not self._hold_in_item(item):
                    main_order_items.append(item.clone())

        self.order_changer.change_order(order.order_id, self._mark_items_as_split)

        new_order = order
        if len(orders) > 0:
            new_order = order.clone()
            new_order.order_id = str(order.order_id) + "-0"

        new_order.items = main_order_items
        orders.insert(0, new_order)

        return orders

    def _handle_hold_item(
        self,
        item,           # type: Item
        order,          # type: ProductionOrder
        orders,         # type: List[ProductionOrder]
        split_index,    # type: int
    ):
        # type: (...) -> Tuple[Any, int]

        stored_orders_states = [x for x in order.state_history if x.state == "STORED"]
        if len(stored_orders_states) == 0:
            return [], split_index

        hold_item_list = self._get_item_list_with_property(item.clone(), self._hold_in_item)
        if not hold_item_list:
            return hold_item_list, split_index

        hold_times_items = {}
        for hold_item in hold_item_list:
            hold_time = hold_item.properties.get("hold", 0)
            hold_time = int(hold_time)
            if hold_time not in hold_times_items:
                hold_times_items[hold_time] = []

            hold_times_items[hold_time].append(hold_item)

        for key in hold_times_items.keys():
            hold_type = "hold_" + str(key)
            for current_order in orders:
                if current_order.type == hold_type:
                    new_items = order.get_cloned_items(
                        allowed_items=hold_times_items[key],
                    )
                    current_order.items.extend(new_items)
                    break
            else:
                new_hold_order = order.clone(hold_times_items[key], split_index)
                new_hold_order.type = "hold_" + str(key)
                orders.append(new_hold_order)
                split_index += 1

        return hold_item_list, split_index

    def _schedule_hold_orders(self, orders):
        first_store_time = self._get_order_display_time(orders)
        if first_store_time is None:
            return

        min_hold_time = None
        filtered_orders = []
        for order in orders:
            hold_time = self._get_schedule_hold_time(order)
            if hold_time is None:
                filtered_orders.append(order)
                continue

            display_time = first_store_time + timedelta(seconds=hold_time)
            if datetime.now() > display_time:
                if not order.display_time:
                    order.display_time = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

                for item in order.items:
                    self._remove_hold_from_items(item)

                filtered_orders.append(order)
            else:
                min_hold_time = display_time if min_hold_time is None or display_time < min_hold_time else min_hold_time

        if min_hold_time is not None:
            wait_time = min_hold_time - datetime.now()
            if wait_time.total_seconds() > 0:
                self.publish_scheduler.schedule_publish(orders[0].order_id, wait_time.seconds)

    def _remove_split_tags(self, orders):
        for order in orders:
            if "-" not in order.order_id:
                continue

            split_id = order.order_id.split("-")[1] + "-"

            for item in order.items:
                self._remove_item_split_tags(item, split_id)

    def _add_fire_hold_tag(self, orders):
        for order in orders:
            main_order = "-0" in order.order_id
            for item in order.items:
                fire_checker = self._fire_in_item if main_order else self._fire_in_item_properties
                items_with_fire = self._get_item_list_with_property(item, fire_checker)
                for fire_item in items_with_fire:
                    fire_item.add_tag("fire")

                hold_checker = self._hold_in_item if main_order else self._hold_in_item_properties
                items_with_hold = self._get_item_list_with_property(item, hold_checker)
                for hold_item in items_with_hold:
                    hold_item.add_tag("hold")

    def _remove_item_split_tags(self, item, split_id):
        regex_filter = "[0-9]-.*"
        new_tags = {}
        if item.tags:
            for tag in item.tags.keys():
                if not re.search(regex_filter, tag):
                    new_tags[tag] = tag
                elif split_id in tag:
                    new_tag = tag.replace(split_id, "")
                    new_tags[new_tag] = new_tag

        item.tags = new_tags

        for son in item.items:
            self._remove_item_split_tags(son, split_id)

    @staticmethod
    def _get_schedule_hold_time(order):
        if not hasattr(order, 'type') or "hold_" not in order.type:
            return None

        hold_time = order.type.replace("hold_", "")
        if hold_time == "-1":
            a_day_seconds = 60 * 60 * 24
            return a_day_seconds

        try:
            hold_time = int(hold_time)
            return hold_time
        except ValueError as _:
            return None

    def _remove_hold_from_items(self, item):
        if "hold" in item.properties:
            del item.properties["hold"]

        for son in item.items:
            self._remove_hold_from_items(son)

    @staticmethod
    def _get_order_display_time(orders):
        stored_orders_states = [x for x in orders[0].state_history if x.state == "STORED"]
        if not stored_orders_states:
            return None

        first_store_state = stored_orders_states[0]
        first_store_time = first_store_state.timestamp.split(".")[0]
        first_store_time = datetime.strptime(first_store_time, DATE_FORMAT)

        return first_store_time

    @staticmethod
    def _get_hold_time(item):
        try:
            hold_time = int(item.properties.get("hold"))
        except Exception as _:
            hold_time = -1

        return hold_time

    def _handle_fire(
        self,
        item,           # type: Item
        order,          # type: ProductionOrder
        orders,         # type: List[ProductionOrder]
        split_index,    # type: int
    ):
        # type: (...) -> Tuple[Any, int]

        fire_item = self._get_item_with_property(item.clone(), self._fire_in_item)
        if fire_item:
            new_fire_order = order.clone([fire_item], split_index)
            orders.append(new_fire_order)
            split_index += 1

        return fire_item, split_index

    def _handle_priority(
        self,
        item,           # type: Item
        order,          # type: ProductionOrder
        orders,         # type: List[ProductionOrder]
        split_index,    # type: int
    ):
        # type: (...) -> Tuple[Any, int]

        priority_item = self._get_item_with_property(item.clone(), self._priority_in_item)
        if priority_item:
            new_priority_order = order.clone([priority_item], split_index)
            orders.append(new_priority_order)
            split_index += 1

        return priority_item, split_index

    def _mark_items_as_split(self, root):
        # type: (Any[ProductionOrder, Item]) -> bool

        need_to_save_order = False
        child_need_to_save_order = False
        for item in root.items:
            need_to_save_order = self._mark_items_as_split(item)

            if "fire" in item.properties and "had_fire" not in item.tags:
                item.add_tag("had_fire")
                need_to_save_order = True

            if "hold" in item.properties and "had_hold" not in item.tags:
                item.add_tag("had_hold")
                need_to_save_order = True

            if "hold" not in item.properties and "had_hold" in item.tags and "clear_hold" not in item.tags:
                item.add_tag("clear_hold")
                need_to_save_order = True

            if "highPriority" in item.properties and "had_high_priority" not in item.tags:
                item.add_tag("had_high_priority")
                need_to_save_order = True

        return need_to_save_order or child_need_to_save_order

    @staticmethod
    def _hold_in_item(item):
        return "hold" in item.properties or "had_hold" in item.tags

    @staticmethod
    def _hold_in_item_properties(item):
        return "hold" in item.properties

    @staticmethod
    def _fire_in_item(item):
        return "fire" in item.properties or "had_fire" in item.tags

    @staticmethod
    def _fire_in_item_properties(item):
        return "fire" in item.properties

    @staticmethod
    def _priority_in_item(item):
        return "highPriority" in item.properties or "had_high_priority" in item.tags

    def _has_fire_hold_or_priority(self, item):
        has_fire = self._fire_in_item(item)
        has_hold = self._hold_in_item(item)
        has_high_priority = self._priority_in_item(item)

        children_has_split = False
        for child in item.items:
            children_has_split = children_has_split or self._has_fire_hold_or_priority(child)
            if children_has_split:
                break

        return has_fire or has_hold or has_high_priority or children_has_split

    def _get_item_with_property(self, item, item_has_property):
        if item_has_property(item):
            return item

        if not item.items:
            return None

        children = []
        for child in item.items:
            tag_child = self._get_item_with_property(child, item_has_property)
            if tag_child:
                children.append(tag_child)

        item.items = children
        return item if children else None

    def _get_item_without_properties(self, item, item_has_properties):
        if any([f(item) for f in item_has_properties]):
            return None

        if not item.items:
            return item

        children = []
        for child in item.items:
            tag_child = self._get_item_without_properties(child, item_has_properties)
            if tag_child:
                children.append(tag_child)

        item.items = children
        return item if children else None

    def _get_item_list_with_property(
        self,
        item,                           # type: Item
        item_has_property_function,     # type: Callable[[Item], Any]
        item_list_with_property=None,   # type: List
    ):
        # type: (...) -> List

        if item_list_with_property is None:
            item_list_with_property = []

        if item_has_property_function(item):
            item_list_with_property.append(item)

        for child in item.items:
            self._get_item_list_with_property(child, item_has_property_function, item_list_with_property)

        return item_list_with_property



from application.service import ChromeUpdaterService


class ChromeUpdaterInteractor(object):
    def __init__(
        self,
        chrome_updater_service  # type: ChromeUpdaterService
    ):
        # type: (...) -> None

        self.chrome_updater_service = chrome_updater_service

    def update_chrome(
        self,
        validate    # type: bool

    ):
        # type: (...) -> str

        return self.chrome_updater_service.update_chrome(
            validate_version=validate,
            restart_hv=True
        )







from __future__ import print_function

import os, sys

# Reminder: Don't put any globals here. They will become unusable once we perform the loader trick below.

class Loader(object):
    _map = dict(str=str,
                bytes=bytes,
                input=input,
                int=int,
                chr=chr,
                range=range,
                filter=filter,
                hex=hex,
                map=map,
                oct=oct,
                zip=zip,
                open=open,
                round=round,
                super=super)

    def __init__(self):
        self._name = __name__
        self._sys = sys
        self.__package__ = __package__
        self.__path__ = __path__
        self.__file__ = __file__
        self.USING_PYTHON2 = True if sys.version_info < (3, 0) else False
        if self.USING_PYTHON2:
            import io, future_builtins
            from future.builtins.newround import newround
            from future.builtins.newsuper import newsuper
            from .utils import python2_input
            self._map = dict(str=unicode, # noqa
                             bytes=str,
                             input=python2_input,
                             int=long, # noqa
                             chr=unichr, # noqa
                             range=xrange, # noqa
                             ascii=future_builtins.ascii,
                             filter=future_builtins.filter,
                             hex=future_builtins.hex,
                             map=future_builtins.map,
                             oct=future_builtins.oct,
                             zip=future_builtins.zip,
                             open=io.open,
                             round=newround,
                             super=newsuper) # noqa
        else:
            self._map['ascii'] = ascii # noqa

        self._map.update(dict(USING_PYTHON2=self.USING_PYTHON2,
                              PY2=self.USING_PYTHON2,
                              PY3=not self.USING_PYTHON2))
        self._manifest = list(self._map)
        self._stdio_wrapped = False

    def __getattr__(self, attr):
        if attr == '__all__':
            if self.USING_PYTHON2:
                return list(self._map)
            else:
                return []
        elif attr == '__package__':
            return self.__package__
        elif attr == '__path__':
            return self.__path__
        elif attr == '__file__':
            return self.__file__
        elif attr == '__loader__':
            return None
        elif attr == '__name__':
            return self._name
        elif attr in self._manifest:
            return self._map[attr]
        else:
            try:
                return self._sys.modules[self._name + '.' + attr]
            except KeyError:
                raise AttributeError("'module' object has no attribute '" + attr + "'")

    def __dir__(self):
        return list(self._map)

    def wrap_stdio(self):
        if self.USING_PYTHON2 and not self._stdio_wrapped:
            import sys, io
            from io import TextIOWrapper

            class StderrTextIOWrapper(TextIOWrapper):
                def write(self, s):
                    if type(s) is unicode: # noqa
                        TextIOWrapper.write(self, s)
                    else:
                        TextIOWrapper.write(self, unicode(s, self.encoding)) # noqa

            original_stdin, original_stdout, original_stderr = sys.stdin, sys.stdout, sys.stderr
            sys.stdin = io.open(sys.stdin.fileno(), encoding=sys.stdin.encoding)
            sys.stdin._original_stream = original_stdin
            sys.stdout = StderrTextIOWrapper(io.FileIO(sys.stdout.fileno(), mode='w'), encoding=sys.stdout.encoding,
                                             line_buffering=True if sys.stdout.isatty() else False)
            sys.stdout._original_stream = original_stdout
            sys.stderr = StderrTextIOWrapper(io.FileIO(sys.stderr.fileno(), mode='w'), encoding=sys.stderr.encoding,
                                             line_buffering=True if sys.stderr.isatty() else False)
            sys.stderr._original_stream = original_stderr
            self._stdio_wrapped = True

    def decode_command_line_args(self):
        import sys
        if self.USING_PYTHON2:
            sys.argv = [i if isinstance(i, unicode) else i.decode(sys.stdin.encoding) for i in sys.argv] # noqa
        return sys.argv

    def encode_command_line_args(self):
        import sys
        if self.USING_PYTHON2:
            sys.argv = [i if isinstance(i, str) else i.encode(sys.stdin.encoding) for i in sys.argv]
        return sys.argv

    def wrap_os_environ_io(self):
        if self.USING_PYTHON2:
            import os, sys
            if getattr(os, '__native_getenv', None):
                return
            native_getenv, native_putenv = os.getenv, os.putenv

            def getenv(varname, value=None):
                v = native_getenv(varname, value)
                if isinstance(v, bytes):
                    v = v.decode(sys.stdin.encoding)
                return v

            def putenv(varname, value):
                if not isinstance(varname, bytes):
                    varname = varname.encode(sys.stdout.encoding)
                if not isinstance(value, bytes):
                    value = value.encode(sys.stdout.encoding)
                native_putenv(varname, value)

            os.getenv, os.putenv = getenv, putenv
            os.__native_getenv, os.__native_putenv = native_getenv, native_putenv

from .utils import RedirectingLoader, Move # noqa

USING_PYTHON2 = True if sys.version_info < (3, 0) else False

MOVES = {'collections': [Move(new_name='UserList', old_module='UserList', old_name='UserList'),
                         Move(new_name='UserDict', old_module='UserDict', old_name='UserDict'),
                         Move(new_name='UserString', old_module='UserString', old_name='UserString')],
         'itertools': [Move(new_name='filterfalse', old_module='itertools', old_name='ifilterfalse'),
                       Move(new_name='zip_longest', old_module='itertools', old_name='izip_longest')],
         'sys': [Move(new_name='intern', old_module='__builtin__', old_name='intern')]}

RENAMES = {'__builtin__': 'builtins',
           'copy_reg': 'copyreg',
           'Queue': 'queue',
           'ConfigParser': 'configparser',
           'repr': 'reprlib',
           'thread': '_thread',
           'dummy_thread': '_dummy_thread'}

if os.name == "nt":
    RENAMES["_winreg"] = "winreg"

for new_module, moves in MOVES.items():
    name = __name__ + '.' + new_module
    sys.modules[name] = RedirectingLoader(new_module)
    for move in moves:
        sys.modules[name]._add_redirect(move)

for old_name, new_name in RENAMES.items():
    sys.modules[__name__ + '.' + new_name] = RedirectingLoader(old_name if USING_PYTHON2 else new_name,
                                                               parent_name=__name__)

loader = Loader()
sys.modules[__name__] = loader



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file or at
# https://developers.google.com/open-source/licenses/bsd

"""A database of Python protocol buffer generated symbols.

SymbolDatabase is the MessageFactory for messages generated at compile time,
and makes it easy to create new instances of a registered type, given only the
type's protocol buffer symbol name.

Example usage::

  db = symbol_database.SymbolDatabase()

  # Register symbols of interest, from one or multiple files.
  db.RegisterFileDescriptor(my_proto_pb2.DESCRIPTOR)
  db.RegisterMessage(my_proto_pb2.MyMessage)
  db.RegisterEnumDescriptor(my_proto_pb2.MyEnum.DESCRIPTOR)

  # The database can be used as a MessageFactory, to generate types based on
  # their name:
  types = db.GetMessages(['my_proto.proto'])
  my_message_instance = types['MyMessage']()

  # The database's underlying descriptor pool can be queried, so it's not
  # necessary to know a type's filename to be able to generate it:
  filename = db.pool.FindFileContainingSymbol('MyMessage')
  my_message_instance = db.GetMessages([filename])['MyMessage']()

  # This functionality is also provided directly via a convenience method:
  my_message_instance = db.GetSymbol('MyMessage')()
"""

import warnings

from google.protobuf.internal import api_implementation
from google.protobuf import descriptor_pool
from google.protobuf import message_factory


class SymbolDatabase():
  """A database of Python generated symbols."""

  # local cache of registered classes.
  _classes = {}

  def __init__(self, pool=None):
    """Initializes a new SymbolDatabase."""
    self.pool = pool or descriptor_pool.DescriptorPool()

  def GetPrototype(self, descriptor):
    warnings.warn('SymbolDatabase.GetPrototype() is deprecated. Please '
                  'use message_factory.GetMessageClass() instead. '
                  'SymbolDatabase.GetPrototype() will be removed soon.')
    return message_factory.GetMessageClass(descriptor)

  def CreatePrototype(self, descriptor):
    warnings.warn('Directly call CreatePrototype() is wrong. Please use '
                  'message_factory.GetMessageClass() instead. '
                  'SymbolDatabase.CreatePrototype() will be removed soon.')
    return message_factory._InternalCreateMessageClass(descriptor)

  def GetMessages(self, files):
    warnings.warn('SymbolDatabase.GetMessages() is deprecated. Please use '
                  'message_factory.GetMessageClassedForFiles() instead. '
                  'SymbolDatabase.GetMessages() will be removed soon.')
    return message_factory.GetMessageClassedForFiles(files, self.pool)

  def RegisterMessage(self, message):
    """Registers the given message type in the local database.

    Calls to GetSymbol() and GetMessages() will return messages registered here.

    Args:
      message: A :class:`google.protobuf.message.Message` subclass (or
        instance); its descriptor will be registered.

    Returns:
      The provided message.
    """

    desc = message.DESCRIPTOR
    self._classes[desc] = message
    self.RegisterMessageDescriptor(desc)
    return message

  def RegisterMessageDescriptor(self, message_descriptor):
    """Registers the given message descriptor in the local database.

    Args:
      message_descriptor (Descriptor): the message descriptor to add.
    """
    if api_implementation.Type() == 'python':
      # pylint: disable=protected-access
      self.pool._AddDescriptor(message_descriptor)

  def RegisterEnumDescriptor(self, enum_descriptor):
    """Registers the given enum descriptor in the local database.

    Args:
      enum_descriptor (EnumDescriptor): The enum descriptor to register.

    Returns:
      EnumDescriptor: The provided descriptor.
    """
    if api_implementation.Type() == 'python':
      # pylint: disable=protected-access
      self.pool._AddEnumDescriptor(enum_descriptor)
    return enum_descriptor

  def RegisterServiceDescriptor(self, service_descriptor):
    """Registers the given service descriptor in the local database.

    Args:
      service_descriptor (ServiceDescriptor): the service descriptor to
        register.
    """
    if api_implementation.Type() == 'python':
      # pylint: disable=protected-access
      self.pool._AddServiceDescriptor(service_descriptor)

  def RegisterFileDescriptor(self, file_descriptor):
    """Registers the given file descriptor in the local database.

    Args:
      file_descriptor (FileDescriptor): The file descriptor to register.
    """
    if api_implementation.Type() == 'python':
      # pylint: disable=protected-access
      self.pool._InternalAddFileDescriptor(file_descriptor)

  def GetSymbol(self, symbol):
    """Tries to find a symbol in the local database.

    Currently, this method only returns message.Message instances, however, if
    may be extended in future to support other symbol types.

    Args:
      symbol (str): a protocol buffer symbol.

    Returns:
      A Python class corresponding to the symbol.

    Raises:
      KeyError: if the symbol could not be found.
    """

    return self._classes[self.pool.FindMessageTypeByName(symbol)]

  def GetMessages(self, files):
    # TODO: Fix the differences with MessageFactory.
    """Gets all registered messages from a specified file.

    Only messages already created and registered will be returned; (this is the
    case for imported _pb2 modules)
    But unlike MessageFactory, this version also returns already defined nested
    messages, but does not register any message extensions.

    Args:
      files (list[str]): The file names to extract messages from.

    Returns:
      A dictionary mapping proto names to the message classes.

    Raises:
      KeyError: if a file could not be found.
    """

    def _GetAllMessages(desc):
      """Walk a message Descriptor and recursively yields all message names."""
      yield desc
      for msg_desc in desc.nested_types:
        for nested_desc in _GetAllMessages(msg_desc):
          yield nested_desc

    result = {}
    for file_name in files:
      file_desc = self.pool.FindFileByName(file_name)
      for msg_desc in file_desc.message_types_by_name.values():
        for desc in _GetAllMessages(msg_desc):
          try:
            result[desc.full_name] = self._classes[desc]
          except KeyError:
            # This descriptor has no registered class, skip it.
            pass
    return result


_DEFAULT = SymbolDatabase(pool=descriptor_pool.Default())


def Default():
  """Returns the default SymbolDatabase."""
  return _DEFAULT



# -*- coding: utf-8 -*-
from application.repository.pos import LoaderUpdaterPOSRepository


class HasPendingLoaderUpdateInteractor(object):
    def __init__(
        self,               # type: HasPendingLoaderUpdateInteractor
        pos_repository,     # type: LoaderUpdaterPOSRepository
    ):
        # type: (...) -> None

        self.pos_repository = pos_repository

    def get_loader_pending_update(
        self,   # type: HasPendingLoaderUpdateInteractor
    ):
        # type: (...) -> bool

        return self.pos_repository.has_pending_loader_to_update()



# engine/url.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php

"""Provides the :class:`~sqlalchemy.engine.url.URL` class which encapsulates
information about a database connection specification.

The URL object is created automatically when
:func:`~sqlalchemy.engine.create_engine` is called with a string
argument; alternatively, the URL is a public-facing construct which can
be used directly and is also accepted directly by ``create_engine()``.
"""

from __future__ import annotations

import collections.abc as collections_abc
import re
from typing import Any
from typing import cast
from typing import Dict
from typing import Iterable
from typing import List
from typing import Mapping
from typing import NamedTuple
from typing import Optional
from typing import overload
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union
from urllib.parse import parse_qsl
from urllib.parse import quote_plus
from urllib.parse import unquote

from .interfaces import Dialect
from .. import exc
from .. import util
from ..dialects import plugins
from ..dialects import registry


class URL(NamedTuple):
    """
    Represent the components of a URL used to connect to a database.

    URLs are typically constructed from a fully formatted URL string, where the
    :func:`.make_url` function is used internally by the
    :func:`_sa.create_engine` function in order to parse the URL string into
    its individual components, which are then used to construct a new
    :class:`.URL` object. When parsing from a formatted URL string, the parsing
    format generally follows
    `RFC-1738 <https://www.ietf.org/rfc/rfc1738.txt>`_, with some exceptions.

    A :class:`_engine.URL` object may also be produced directly, either by
    using the :func:`.make_url` function with a fully formed URL string, or
    by using the :meth:`_engine.URL.create` constructor in order
    to construct a :class:`_engine.URL` programmatically given individual
    fields. The resulting :class:`.URL` object may be passed directly to
    :func:`_sa.create_engine` in place of a string argument, which will bypass
    the usage of :func:`.make_url` within the engine's creation process.

    .. versionchanged:: 1.4

        The :class:`_engine.URL` object is now an immutable object.  To
        create a URL, use the :func:`_engine.make_url` or
        :meth:`_engine.URL.create` function / method.  To modify
        a :class:`_engine.URL`, use methods like
        :meth:`_engine.URL.set` and
        :meth:`_engine.URL.update_query_dict` to return a new
        :class:`_engine.URL` object with modifications.   See notes for this
        change at :ref:`change_5526`.

    .. seealso::

        :ref:`database_urls`

    :class:`_engine.URL` contains the following attributes:

    * :attr:`_engine.URL.drivername`: database backend and driver name, such as
      ``postgresql+psycopg2``
    * :attr:`_engine.URL.username`: username string
    * :attr:`_engine.URL.password`: password string
    * :attr:`_engine.URL.host`: string hostname
    * :attr:`_engine.URL.port`: integer port number
    * :attr:`_engine.URL.database`: string database name
    * :attr:`_engine.URL.query`: an immutable mapping representing the query
      string.  contains strings for keys and either strings or tuples of
      strings for values.


    """

    drivername: str
    """database backend and driver name, such as
    ``postgresql+psycopg2``

    """

    username: Optional[str]
    "username string"

    password: Optional[str]
    """password, which is normally a string but may also be any
    object that has a ``__str__()`` method."""

    host: Optional[str]
    """hostname or IP number.  May also be a data source name for some
    drivers."""

    port: Optional[int]
    """integer port number"""

    database: Optional[str]
    """database name"""

    query: util.immutabledict[str, Union[Tuple[str, ...], str]]
    """an immutable mapping representing the query string.  contains strings
       for keys and either strings or tuples of strings for values, e.g.::

            >>> from sqlalchemy.engine import make_url
            >>> url = make_url("postgresql+psycopg2://user:pass@host/dbname?alt_host=host1&alt_host=host2&ssl_cipher=%2Fpath%2Fto%2Fcrt")
            >>> url.query
            immutabledict({'alt_host': ('host1', 'host2'), 'ssl_cipher': '/path/to/crt'})

         To create a mutable copy of this mapping, use the ``dict`` constructor::

            mutable_query_opts = dict(url.query)

       .. seealso::

          :attr:`_engine.URL.normalized_query` - normalizes all values into sequences
          for consistent processing

          Methods for altering the contents of :attr:`_engine.URL.query`:

          :meth:`_engine.URL.update_query_dict`

          :meth:`_engine.URL.update_query_string`

          :meth:`_engine.URL.update_query_pairs`

          :meth:`_engine.URL.difference_update_query`

    """  # noqa: E501

    @classmethod
    def create(
        cls,
        drivername: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        database: Optional[str] = None,
        query: Mapping[str, Union[Sequence[str], str]] = util.EMPTY_DICT,
    ) -> URL:
        """Create a new :class:`_engine.URL` object.

        .. seealso::

            :ref:`database_urls`

        :param drivername: the name of the database backend. This name will
          correspond to a module in sqlalchemy/databases or a third party
          plug-in.
        :param username: The user name.
        :param password: database password.  Is typically a string, but may
          also be an object that can be stringified with ``str()``.

          .. note::  A password-producing object will be stringified only
             **once** per :class:`_engine.Engine` object.  For dynamic password
             generation per connect, see :ref:`engines_dynamic_tokens`.

        :param host: The name of the host.
        :param port: The port number.
        :param database: The database name.
        :param query: A dictionary of string keys to string values to be passed
          to the dialect and/or the DBAPI upon connect.   To specify non-string
          parameters to a Python DBAPI directly, use the
          :paramref:`_sa.create_engine.connect_args` parameter to
          :func:`_sa.create_engine`.   See also
          :attr:`_engine.URL.normalized_query` for a dictionary that is
          consistently string->list of string.
        :return: new :class:`_engine.URL` object.

        .. versionadded:: 1.4

            The :class:`_engine.URL` object is now an **immutable named
            tuple**.  In addition, the ``query`` dictionary is also immutable.
            To create a URL, use the :func:`_engine.url.make_url` or
            :meth:`_engine.URL.create` function/ method.  To modify a
            :class:`_engine.URL`, use the :meth:`_engine.URL.set` and
            :meth:`_engine.URL.update_query` methods.

        """

        return cls(
            cls._assert_str(drivername, "drivername"),
            cls._assert_none_str(username, "username"),
            password,
            cls._assert_none_str(host, "host"),
            cls._assert_port(port),
            cls._assert_none_str(database, "database"),
            cls._str_dict(query),
        )

    @classmethod
    def _assert_port(cls, port: Optional[int]) -> Optional[int]:
        if port is None:
            return None
        try:
            return int(port)
        except TypeError:
            raise TypeError("Port argument must be an integer or None")

    @classmethod
    def _assert_str(cls, v: str, paramname: str) -> str:
        if not isinstance(v, str):
            raise TypeError("%s must be a string" % paramname)
        return v

    @classmethod
    def _assert_none_str(
        cls, v: Optional[str], paramname: str
    ) -> Optional[str]:
        if v is None:
            return v

        return cls._assert_str(v, paramname)

    @classmethod
    def _str_dict(
        cls,
        dict_: Optional[
            Union[
                Sequence[Tuple[str, Union[Sequence[str], str]]],
                Mapping[str, Union[Sequence[str], str]],
            ]
        ],
    ) -> util.immutabledict[str, Union[Tuple[str, ...], str]]:
        if dict_ is None:
            return util.EMPTY_DICT

        @overload
        def _assert_value(
            val: str,
        ) -> str:
            ...

        @overload
        def _assert_value(
            val: Sequence[str],
        ) -> Union[str, Tuple[str, ...]]:
            ...

        def _assert_value(
            val: Union[str, Sequence[str]],
        ) -> Union[str, Tuple[str, ...]]:
            if isinstance(val, str):
                return val
            elif isinstance(val, collections_abc.Sequence):
                return tuple(_assert_value(elem) for elem in val)
            else:
                raise TypeError(
                    "Query dictionary values must be strings or "
                    "sequences of strings"
                )

        def _assert_str(v: str) -> str:
            if not isinstance(v, str):
                raise TypeError("Query dictionary keys must be strings")
            return v

        dict_items: Iterable[Tuple[str, Union[Sequence[str], str]]]
        if isinstance(dict_, collections_abc.Sequence):
            dict_items = dict_
        else:
            dict_items = dict_.items()

        return util.immutabledict(
            {
                _assert_str(key): _assert_value(
                    value,
                )
                for key, value in dict_items
            }
        )

    def set(
        self,
        drivername: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        database: Optional[str] = None,
        query: Optional[Mapping[str, Union[Sequence[str], str]]] = None,
    ) -> URL:
        """return a new :class:`_engine.URL` object with modifications.

        Values are used if they are non-None.  To set a value to ``None``
        explicitly, use the :meth:`_engine.URL._replace` method adapted
        from ``namedtuple``.

        :param drivername: new drivername
        :param username: new username
        :param password: new password
        :param host: new hostname
        :param port: new port
        :param query: new query parameters, passed a dict of string keys
         referring to string or sequence of string values.  Fully
         replaces the previous list of arguments.

        :return: new :class:`_engine.URL` object.

        .. versionadded:: 1.4

        .. seealso::

            :meth:`_engine.URL.update_query_dict`

        """

        kw: Dict[str, Any] = {}
        if drivername is not None:
            kw["drivername"] = drivername
        if username is not None:
            kw["username"] = username
        if password is not None:
            kw["password"] = password
        if host is not None:
            kw["host"] = host
        if port is not None:
            kw["port"] = port
        if database is not None:
            kw["database"] = database
        if query is not None:
            kw["query"] = query

        return self._assert_replace(**kw)

    def _assert_replace(self, **kw: Any) -> URL:
        """argument checks before calling _replace()"""

        if "drivername" in kw:
            self._assert_str(kw["drivername"], "drivername")
        for name in "username", "host", "database":
            if name in kw:
                self._assert_none_str(kw[name], name)
        if "port" in kw:
            self._assert_port(kw["port"])
        if "query" in kw:
            kw["query"] = self._str_dict(kw["query"])

        return self._replace(**kw)

    def update_query_string(
        self, query_string: str, append: bool = False
    ) -> URL:
        """Return a new :class:`_engine.URL` object with the :attr:`_engine.URL.query`
        parameter dictionary updated by the given query string.

        E.g.::

            >>> from sqlalchemy.engine import make_url
            >>> url = make_url("postgresql+psycopg2://user:pass@host/dbname")
            >>> url = url.update_query_string("alt_host=host1&alt_host=host2&ssl_cipher=%2Fpath%2Fto%2Fcrt")
            >>> str(url)
            'postgresql+psycopg2://user:pass@host/dbname?alt_host=host1&alt_host=host2&ssl_cipher=%2Fpath%2Fto%2Fcrt'

        :param query_string: a URL escaped query string, not including the
         question mark.

        :param append: if True, parameters in the existing query string will
         not be removed; new parameters will be in addition to those present.
         If left at its default of False, keys present in the given query
         parameters will replace those of the existing query string.

        .. versionadded:: 1.4

        .. seealso::

            :attr:`_engine.URL.query`

            :meth:`_engine.URL.update_query_dict`

        """  # noqa: E501
        return self.update_query_pairs(parse_qsl(query_string), append=append)

    def update_query_pairs(
        self,
        key_value_pairs: Iterable[Tuple[str, Union[str, List[str]]]],
        append: bool = False,
    ) -> URL:
        """Return a new :class:`_engine.URL` object with the
        :attr:`_engine.URL.query`
        parameter dictionary updated by the given sequence of key/value pairs

        E.g.::

            >>> from sqlalchemy.engine import make_url
            >>> url = make_url("postgresql+psycopg2://user:pass@host/dbname")
            >>> url = url.update_query_pairs([("alt_host", "host1"), ("alt_host", "host2"), ("ssl_cipher", "/path/to/crt")])
            >>> str(url)
            'postgresql+psycopg2://user:pass@host/dbname?alt_host=host1&alt_host=host2&ssl_cipher=%2Fpath%2Fto%2Fcrt'

        :param key_value_pairs: A sequence of tuples containing two strings
         each.

        :param append: if True, parameters in the existing query string will
         not be removed; new parameters will be in addition to those present.
         If left at its default of False, keys present in the given query
         parameters will replace those of the existing query string.

        .. versionadded:: 1.4

        .. seealso::

            :attr:`_engine.URL.query`

            :meth:`_engine.URL.difference_update_query`

            :meth:`_engine.URL.set`

        """  # noqa: E501

        existing_query = self.query
        new_keys: Dict[str, Union[str, List[str]]] = {}

        for key, value in key_value_pairs:
            if key in new_keys:
                new_keys[key] = util.to_list(new_keys[key])
                cast("List[str]", new_keys[key]).append(cast(str, value))
            else:
                new_keys[key] = (
                    list(value) if isinstance(value, (list, tuple)) else value
                )

        new_query: Mapping[str, Union[str, Sequence[str]]]
        if append:
            new_query = {}

            for k in new_keys:
                if k in existing_query:
                    new_query[k] = tuple(
                        util.to_list(existing_query[k])
                        + util.to_list(new_keys[k])
                    )
                else:
                    new_query[k] = new_keys[k]

            new_query.update(
                {
                    k: existing_query[k]
                    for k in set(existing_query).difference(new_keys)
                }
            )
        else:
            new_query = self.query.union(
                {
                    k: tuple(v) if isinstance(v, list) else v
                    for k, v in new_keys.items()
                }
            )
        return self.set(query=new_query)

    def update_query_dict(
        self,
        query_parameters: Mapping[str, Union[str, List[str]]],
        append: bool = False,
    ) -> URL:
        """Return a new :class:`_engine.URL` object with the
        :attr:`_engine.URL.query` parameter dictionary updated by the given
        dictionary.

        The dictionary typically contains string keys and string values.
        In order to represent a query parameter that is expressed multiple
        times, pass a sequence of string values.

        E.g.::


            >>> from sqlalchemy.engine import make_url
            >>> url = make_url("postgresql+psycopg2://user:pass@host/dbname")
            >>> url = url.update_query_dict({"alt_host": ["host1", "host2"], "ssl_cipher": "/path/to/crt"})
            >>> str(url)
            'postgresql+psycopg2://user:pass@host/dbname?alt_host=host1&alt_host=host2&ssl_cipher=%2Fpath%2Fto%2Fcrt'


        :param query_parameters: A dictionary with string keys and values
         that are either strings, or sequences of strings.

        :param append: if True, parameters in the existing query string will
         not be removed; new parameters will be in addition to those present.
         If left at its default of False, keys present in the given query
         parameters will replace those of the existing query string.


        .. versionadded:: 1.4

        .. seealso::

            :attr:`_engine.URL.query`

            :meth:`_engine.URL.update_query_string`

            :meth:`_engine.URL.update_query_pairs`

            :meth:`_engine.URL.difference_update_query`

            :meth:`_engine.URL.set`

        """  # noqa: E501
        return self.update_query_pairs(query_parameters.items(), append=append)

    def difference_update_query(self, names: Iterable[str]) -> URL:
        """
        Remove the given names from the :attr:`_engine.URL.query` dictionary,
        returning the new :class:`_engine.URL`.

        E.g.::

            url = url.difference_update_query(['foo', 'bar'])

        Equivalent to using :meth:`_engine.URL.set` as follows::

            url = url.set(
                query={
                    key: url.query[key]
                    for key in set(url.query).difference(['foo', 'bar'])
                }
            )

        .. versionadded:: 1.4

        .. seealso::

            :attr:`_engine.URL.query`

            :meth:`_engine.URL.update_query_dict`

            :meth:`_engine.URL.set`

        """

        if not set(names).intersection(self.query):
            return self

        return URL(
            self.drivername,
            self.username,
            self.password,
            self.host,
            self.port,
            self.database,
            util.immutabledict(
                {
                    key: self.query[key]
                    for key in set(self.query).difference(names)
                }
            ),
        )

    @property
    def normalized_query(self) -> Mapping[str, Sequence[str]]:
        """Return the :attr:`_engine.URL.query` dictionary with values normalized
        into sequences.

        As the :attr:`_engine.URL.query` dictionary may contain either
        string values or sequences of string values to differentiate between
        parameters that are specified multiple times in the query string,
        code that needs to handle multiple parameters generically will wish
        to use this attribute so that all parameters present are presented
        as sequences.   Inspiration is from Python's ``urllib.parse.parse_qs``
        function.  E.g.::


            >>> from sqlalchemy.engine import make_url
            >>> url = make_url("postgresql+psycopg2://user:pass@host/dbname?alt_host=host1&alt_host=host2&ssl_cipher=%2Fpath%2Fto%2Fcrt")
            >>> url.query
            immutabledict({'alt_host': ('host1', 'host2'), 'ssl_cipher': '/path/to/crt'})
            >>> url.normalized_query
            immutabledict({'alt_host': ('host1', 'host2'), 'ssl_cipher': ('/path/to/crt',)})

        """  # noqa: E501

        return util.immutabledict(
            {
                k: (v,) if not isinstance(v, tuple) else v
                for k, v in self.query.items()
            }
        )

    @util.deprecated(
        "1.4",
        "The :meth:`_engine.URL.__to_string__ method is deprecated and will "
        "be removed in a future release.  Please use the "
        ":meth:`_engine.URL.render_as_string` method.",
    )
    def __to_string__(self, hide_password: bool = True) -> str:
        """Render this :class:`_engine.URL` object as a string.

        :param hide_password: Defaults to True.   The password is not shown
         in the string unless this is set to False.

        """
        return self.render_as_string(hide_password=hide_password)

    def render_as_string(self, hide_password: bool = True) -> str:
        """Render this :class:`_engine.URL` object as a string.

        This method is used when the ``__str__()`` or ``__repr__()``
        methods are used.   The method directly includes additional options.

        :param hide_password: Defaults to True.   The password is not shown
         in the string unless this is set to False.

        """
        s = self.drivername + "://"
        if self.username is not None:
            s += _sqla_url_quote(self.username)
            if self.password is not None:
                s += ":" + (
                    "***"
                    if hide_password
                    else _sqla_url_quote(str(self.password))
                )
            s += "@"
        if self.host is not None:
            if ":" in self.host:
                s += "[%s]" % self.host
            else:
                s += self.host
        if self.port is not None:
            s += ":" + str(self.port)
        if self.database is not None:
            s += "/" + self.database
        if self.query:
            keys = list(self.query)
            keys.sort()
            s += "?" + "&".join(
                "%s=%s" % (quote_plus(k), quote_plus(element))
                for k in keys
                for element in util.to_list(self.query[k])
            )
        return s

    def __repr__(self) -> str:
        return self.render_as_string()

    def __copy__(self) -> URL:
        return self.__class__.create(
            self.drivername,
            self.username,
            self.password,
            self.host,
            self.port,
            self.database,
            # note this is an immutabledict of str-> str / tuple of str,
            # also fully immutable.  does not require deepcopy
            self.query,
        )

    def __deepcopy__(self, memo: Any) -> URL:
        return self.__copy__()

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, URL)
            and self.drivername == other.drivername
            and self.username == other.username
            and self.password == other.password
            and self.host == other.host
            and self.database == other.database
            and self.query == other.query
            and self.port == other.port
        )

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def get_backend_name(self) -> str:
        """Return the backend name.

        This is the name that corresponds to the database backend in
        use, and is the portion of the :attr:`_engine.URL.drivername`
        that is to the left of the plus sign.

        """
        if "+" not in self.drivername:
            return self.drivername
        else:
            return self.drivername.split("+")[0]

    def get_driver_name(self) -> str:
        """Return the backend name.

        This is the name that corresponds to the DBAPI driver in
        use, and is the portion of the :attr:`_engine.URL.drivername`
        that is to the right of the plus sign.

        If the :attr:`_engine.URL.drivername` does not include a plus sign,
        then the default :class:`_engine.Dialect` for this :class:`_engine.URL`
        is imported in order to get the driver name.

        """

        if "+" not in self.drivername:
            return self.get_dialect().driver
        else:
            return self.drivername.split("+")[1]

    def _instantiate_plugins(
        self, kwargs: Mapping[str, Any]
    ) -> Tuple[URL, List[Any], Dict[str, Any]]:
        plugin_names = util.to_list(self.query.get("plugin", ()))
        plugin_names += kwargs.get("plugins", [])

        kwargs = dict(kwargs)

        loaded_plugins = [
            plugins.load(plugin_name)(self, kwargs)
            for plugin_name in plugin_names
        ]

        u = self.difference_update_query(["plugin", "plugins"])

        for plugin in loaded_plugins:
            new_u = plugin.update_url(u)
            if new_u is not None:
                u = new_u

        kwargs.pop("plugins", None)

        return u, loaded_plugins, kwargs

    def _get_entrypoint(self) -> Type[Dialect]:
        """Return the "entry point" dialect class.

        This is normally the dialect itself except in the case when the
        returned class implements the get_dialect_cls() method.

        """
        if "+" not in self.drivername:
            name = self.drivername
        else:
            name = self.drivername.replace("+", ".")
        cls = registry.load(name)
        # check for legacy dialects that
        # would return a module with 'dialect' as the
        # actual class
        if (
            hasattr(cls, "dialect")
            and isinstance(cls.dialect, type)
            and issubclass(cls.dialect, Dialect)
        ):
            return cls.dialect
        else:
            return cast("Type[Dialect]", cls)

    def get_dialect(self, _is_async: bool = False) -> Type[Dialect]:
        """Return the SQLAlchemy :class:`_engine.Dialect` class corresponding
        to this URL's driver name.

        """
        entrypoint = self._get_entrypoint()
        if _is_async:
            dialect_cls = entrypoint.get_async_dialect_cls(self)
        else:
            dialect_cls = entrypoint.get_dialect_cls(self)
        return dialect_cls

    def translate_connect_args(
        self, names: Optional[List[str]] = None, **kw: Any
    ) -> Dict[str, Any]:
        r"""Translate url attributes into a dictionary of connection arguments.

        Returns attributes of this url (`host`, `database`, `username`,
        `password`, `port`) as a plain dictionary.  The attribute names are
        used as the keys by default.  Unset or false attributes are omitted
        from the final dictionary.

        :param \**kw: Optional, alternate key names for url attributes.

        :param names: Deprecated.  Same purpose as the keyword-based alternate
            names, but correlates the name to the original positionally.
        """

        if names is not None:
            util.warn_deprecated(
                "The `URL.translate_connect_args.name`s parameter is "
                "deprecated. Please pass the "
                "alternate names as kw arguments.",
                "1.4",
            )

        translated = {}
        attribute_names = ["host", "database", "username", "password", "port"]
        for sname in attribute_names:
            if names:
                name = names.pop(0)
            elif sname in kw:
                name = kw[sname]
            else:
                name = sname
            if name is not None and getattr(self, sname, False):
                if sname == "password":
                    translated[name] = str(getattr(self, sname))
                else:
                    translated[name] = getattr(self, sname)

        return translated


def make_url(name_or_url: Union[str, URL]) -> URL:
    """Given a string, produce a new URL instance.

    The format of the URL generally follows `RFC-1738
    <https://www.ietf.org/rfc/rfc1738.txt>`_, with some exceptions, including
    that underscores, and not dashes or periods, are accepted within the
    "scheme" portion.

    If a :class:`.URL` object is passed, it is returned as is.

    .. seealso::

        :ref:`database_urls`

    """

    if isinstance(name_or_url, str):
        return _parse_url(name_or_url)
    elif not isinstance(name_or_url, URL) and not hasattr(
        name_or_url, "_sqla_is_testing_if_this_is_a_mock_object"
    ):
        raise exc.ArgumentError(
            f"Expected string or URL object, got {name_or_url!r}"
        )
    else:
        return name_or_url


def _parse_url(name: str) -> URL:
    pattern = re.compile(
        r"""
            (?P<name>[\w\+]+)://
            (?:
                (?P<username>[^:/]*)
                (?::(?P<password>[^@]*))?
            @)?
            (?:
                (?:
                    \[(?P<ipv6host>[^/\?]+)\] |
                    (?P<ipv4host>[^/:\?]+)
                )?
                (?::(?P<port>[^/\?]*))?
            )?
            (?:/(?P<database>[^\?]*))?
            (?:\?(?P<query>.*))?
            """,
        re.X,
    )

    m = pattern.match(name)
    if m is not None:
        components = m.groupdict()
        query: Optional[Dict[str, Union[str, List[str]]]]
        if components["query"] is not None:
            query = {}

            for key, value in parse_qsl(components["query"]):
                if key in query:
                    query[key] = util.to_list(query[key])
                    cast("List[str]", query[key]).append(value)
                else:
                    query[key] = value
        else:
            query = None
        components["query"] = query

        if components["username"] is not None:
            components["username"] = _sqla_url_unquote(components["username"])

        if components["password"] is not None:
            components["password"] = _sqla_url_unquote(components["password"])

        ipv4host = components.pop("ipv4host")
        ipv6host = components.pop("ipv6host")
        components["host"] = ipv4host or ipv6host
        name = components.pop("name")

        if components["port"]:
            components["port"] = int(components["port"])

        return URL.create(name, **components)  # type: ignore

    else:
        raise exc.ArgumentError(
            "Could not parse SQLAlchemy URL from string '%s'" % name
        )


def _sqla_url_quote(text: str) -> str:
    return re.sub(r"[:@/]", lambda m: "%%%X" % ord(m.group(0)), text)


_sqla_url_unquote = unquote



import io

from ...migration import MigrationContext
from ...testing import assert_raises
from ...testing import config
from ...testing import eq_
from ...testing import is_
from ...testing import is_false
from ...testing import is_not_
from ...testing import is_true
from ...testing import ne_
from ...testing.fixtures import TestBase


class MigrationTransactionTest(TestBase):
    __backend__ = True

    conn = None

    def _fixture(self, opts):
        self.conn = conn = config.db.connect()

        if opts.get("as_sql", False):
            self.context = MigrationContext.configure(
                dialect=conn.dialect, opts=opts
            )
            self.context.output_buffer = (
                self.context.impl.output_buffer
            ) = io.StringIO()
        else:
            self.context = MigrationContext.configure(
                connection=conn, opts=opts
            )
        return self.context

    def teardown_method(self):
        if self.conn:
            self.conn.close()

    def test_proxy_transaction_rollback(self):
        context = self._fixture(
            {"transaction_per_migration": True, "transactional_ddl": True}
        )

        is_false(self.conn.in_transaction())
        proxy = context.begin_transaction(_per_migration=True)
        is_true(self.conn.in_transaction())
        proxy.rollback()
        is_false(self.conn.in_transaction())

    def test_proxy_transaction_commit(self):
        context = self._fixture(
            {"transaction_per_migration": True, "transactional_ddl": True}
        )
        proxy = context.begin_transaction(_per_migration=True)
        is_true(self.conn.in_transaction())
        proxy.commit()
        is_false(self.conn.in_transaction())

    def test_proxy_transaction_contextmanager_commit(self):
        context = self._fixture(
            {"transaction_per_migration": True, "transactional_ddl": True}
        )
        proxy = context.begin_transaction(_per_migration=True)
        is_true(self.conn.in_transaction())
        with proxy:
            pass
        is_false(self.conn.in_transaction())

    def test_proxy_transaction_contextmanager_rollback(self):
        context = self._fixture(
            {"transaction_per_migration": True, "transactional_ddl": True}
        )
        proxy = context.begin_transaction(_per_migration=True)
        is_true(self.conn.in_transaction())

        def go():
            with proxy:
                raise Exception("hi")

        assert_raises(Exception, go)
        is_false(self.conn.in_transaction())

    def test_proxy_transaction_contextmanager_explicit_rollback(self):
        context = self._fixture(
            {"transaction_per_migration": True, "transactional_ddl": True}
        )
        proxy = context.begin_transaction(_per_migration=True)
        is_true(self.conn.in_transaction())

        with proxy:
            is_true(self.conn.in_transaction())
            proxy.rollback()
            is_false(self.conn.in_transaction())

        is_false(self.conn.in_transaction())

    def test_proxy_transaction_contextmanager_explicit_commit(self):
        context = self._fixture(
            {"transaction_per_migration": True, "transactional_ddl": True}
        )
        proxy = context.begin_transaction(_per_migration=True)
        is_true(self.conn.in_transaction())

        with proxy:
            is_true(self.conn.in_transaction())
            proxy.commit()
            is_false(self.conn.in_transaction())

        is_false(self.conn.in_transaction())

    def test_transaction_per_migration_transactional_ddl(self):
        context = self._fixture(
            {"transaction_per_migration": True, "transactional_ddl": True}
        )

        is_false(self.conn.in_transaction())

        with context.begin_transaction():
            is_false(self.conn.in_transaction())
            with context.begin_transaction(_per_migration=True):
                is_true(self.conn.in_transaction())

            is_false(self.conn.in_transaction())
        is_false(self.conn.in_transaction())

    def test_transaction_per_migration_non_transactional_ddl(self):
        context = self._fixture(
            {"transaction_per_migration": True, "transactional_ddl": False}
        )

        is_false(self.conn.in_transaction())

        with context.begin_transaction():
            is_false(self.conn.in_transaction())
            with context.begin_transaction(_per_migration=True):
                is_true(self.conn.in_transaction())

            is_false(self.conn.in_transaction())
        is_false(self.conn.in_transaction())

    def test_transaction_per_all_transactional_ddl(self):
        context = self._fixture({"transactional_ddl": True})

        is_false(self.conn.in_transaction())

        with context.begin_transaction():
            is_true(self.conn.in_transaction())
            with context.begin_transaction(_per_migration=True):
                is_true(self.conn.in_transaction())

            is_true(self.conn.in_transaction())
        is_false(self.conn.in_transaction())

    def test_transaction_per_all_non_transactional_ddl(self):
        context = self._fixture({"transactional_ddl": False})

        is_false(self.conn.in_transaction())

        with context.begin_transaction():
            is_false(self.conn.in_transaction())
            with context.begin_transaction(_per_migration=True):
                is_true(self.conn.in_transaction())

            is_false(self.conn.in_transaction())
        is_false(self.conn.in_transaction())

    def test_transaction_per_all_sqlmode(self):
        context = self._fixture({"as_sql": True})

        context.execute("step 1")
        with context.begin_transaction():
            context.execute("step 2")
            with context.begin_transaction(_per_migration=True):
                context.execute("step 3")

            context.execute("step 4")
        context.execute("step 5")

        if context.impl.transactional_ddl:
            self._assert_impl_steps(
                "step 1",
                "BEGIN",
                "step 2",
                "step 3",
                "step 4",
                "COMMIT",
                "step 5",
            )
        else:
            self._assert_impl_steps(
                "step 1", "step 2", "step 3", "step 4", "step 5"
            )

    def test_transaction_per_migration_sqlmode(self):
        context = self._fixture(
            {"as_sql": True, "transaction_per_migration": True}
        )

        context.execute("step 1")
        with context.begin_transaction():
            context.execute("step 2")
            with context.begin_transaction(_per_migration=True):
                context.execute("step 3")

            context.execute("step 4")
        context.execute("step 5")

        if context.impl.transactional_ddl:
            self._assert_impl_steps(
                "step 1",
                "step 2",
                "BEGIN",
                "step 3",
                "COMMIT",
                "step 4",
                "step 5",
            )
        else:
            self._assert_impl_steps(
                "step 1", "step 2", "step 3", "step 4", "step 5"
            )

    @config.requirements.autocommit_isolation
    def test_autocommit_block(self):
        context = self._fixture({"transaction_per_migration": True})

        is_false(self.conn.in_transaction())

        with context.begin_transaction():
            is_false(self.conn.in_transaction())
            with context.begin_transaction(_per_migration=True):
                is_true(self.conn.in_transaction())

                with context.autocommit_block():
                    # in 1.x, self.conn is separate due to the
                    # execution_options call.  however for future they are the
                    # same connection and there is a "transaction" block
                    # despite autocommit
                    if self.is_sqlalchemy_future:
                        is_(context.connection, self.conn)
                    else:
                        is_not_(context.connection, self.conn)
                        is_false(self.conn.in_transaction())

                    eq_(
                        context.connection._execution_options[
                            "isolation_level"
                        ],
                        "AUTOCOMMIT",
                    )

                ne_(
                    context.connection._execution_options.get(
                        "isolation_level", None
                    ),
                    "AUTOCOMMIT",
                )
                is_true(self.conn.in_transaction())

            is_false(self.conn.in_transaction())
        is_false(self.conn.in_transaction())

    @config.requirements.autocommit_isolation
    def test_autocommit_block_no_transaction(self):
        context = self._fixture({"transaction_per_migration": True})

        is_false(self.conn.in_transaction())

        with context.autocommit_block():
            is_true(context.connection.in_transaction())

            # in 1.x, self.conn is separate due to the execution_options
            # call.  however for future they are the same connection and there
            # is a "transaction" block despite autocommit
            if self.is_sqlalchemy_future:
                is_(context.connection, self.conn)
            else:
                is_not_(context.connection, self.conn)
                is_false(self.conn.in_transaction())

            eq_(
                context.connection._execution_options["isolation_level"],
                "AUTOCOMMIT",
            )

        ne_(
            context.connection._execution_options.get("isolation_level", None),
            "AUTOCOMMIT",
        )

        is_false(self.conn.in_transaction())

    def test_autocommit_block_transactional_ddl_sqlmode(self):
        context = self._fixture(
            {
                "transaction_per_migration": True,
                "transactional_ddl": True,
                "as_sql": True,
            }
        )

        with context.begin_transaction():
            context.execute("step 1")
            with context.begin_transaction(_per_migration=True):
                context.execute("step 2")

                with context.autocommit_block():
                    context.execute("step 3")

                context.execute("step 4")

            context.execute("step 5")

        self._assert_impl_steps(
            "step 1",
            "BEGIN",
            "step 2",
            "COMMIT",
            "step 3",
            "BEGIN",
            "step 4",
            "COMMIT",
            "step 5",
        )

    def test_autocommit_block_nontransactional_ddl_sqlmode(self):
        context = self._fixture(
            {
                "transaction_per_migration": True,
                "transactional_ddl": False,
                "as_sql": True,
            }
        )

        with context.begin_transaction():
            context.execute("step 1")
            with context.begin_transaction(_per_migration=True):
                context.execute("step 2")

                with context.autocommit_block():
                    context.execute("step 3")

                context.execute("step 4")

            context.execute("step 5")

        self._assert_impl_steps(
            "step 1", "step 2", "step 3", "step 4", "step 5"
        )

    def _assert_impl_steps(self, *steps):
        to_check = self.context.output_buffer.getvalue()

        self.context.impl.output_buffer = buf = io.StringIO()
        for step in steps:
            if step == "BEGIN":
                self.context.impl.emit_begin()
            elif step == "COMMIT":
                self.context.impl.emit_commit()
            else:
                self.context.impl._exec(step)

        eq_(to_check, buf.getvalue())



data = (
'bbess',    # 0x00
'bbeng',    # 0x01
'bbej',    # 0x02
'bbec',    # 0x03
'bbek',    # 0x04
'bbet',    # 0x05
'bbep',    # 0x06
'bbeh',    # 0x07
'bbyeo',    # 0x08
'bbyeog',    # 0x09
'bbyeogg',    # 0x0a
'bbyeogs',    # 0x0b
'bbyeon',    # 0x0c
'bbyeonj',    # 0x0d
'bbyeonh',    # 0x0e
'bbyeod',    # 0x0f
'bbyeol',    # 0x10
'bbyeolg',    # 0x11
'bbyeolm',    # 0x12
'bbyeolb',    # 0x13
'bbyeols',    # 0x14
'bbyeolt',    # 0x15
'bbyeolp',    # 0x16
'bbyeolh',    # 0x17
'bbyeom',    # 0x18
'bbyeob',    # 0x19
'bbyeobs',    # 0x1a
'bbyeos',    # 0x1b
'bbyeoss',    # 0x1c
'bbyeong',    # 0x1d
'bbyeoj',    # 0x1e
'bbyeoc',    # 0x1f
'bbyeok',    # 0x20
'bbyeot',    # 0x21
'bbyeop',    # 0x22
'bbyeoh',    # 0x23
'bbye',    # 0x24
'bbyeg',    # 0x25
'bbyegg',    # 0x26
'bbyegs',    # 0x27
'bbyen',    # 0x28
'bbyenj',    # 0x29
'bbyenh',    # 0x2a
'bbyed',    # 0x2b
'bbyel',    # 0x2c
'bbyelg',    # 0x2d
'bbyelm',    # 0x2e
'bbyelb',    # 0x2f
'bbyels',    # 0x30
'bbyelt',    # 0x31
'bbyelp',    # 0x32
'bbyelh',    # 0x33
'bbyem',    # 0x34
'bbyeb',    # 0x35
'bbyebs',    # 0x36
'bbyes',    # 0x37
'bbyess',    # 0x38
'bbyeng',    # 0x39
'bbyej',    # 0x3a
'bbyec',    # 0x3b
'bbyek',    # 0x3c
'bbyet',    # 0x3d
'bbyep',    # 0x3e
'bbyeh',    # 0x3f
'bbo',    # 0x40
'bbog',    # 0x41
'bbogg',    # 0x42
'bbogs',    # 0x43
'bbon',    # 0x44
'bbonj',    # 0x45
'bbonh',    # 0x46
'bbod',    # 0x47
'bbol',    # 0x48
'bbolg',    # 0x49
'bbolm',    # 0x4a
'bbolb',    # 0x4b
'bbols',    # 0x4c
'bbolt',    # 0x4d
'bbolp',    # 0x4e
'bbolh',    # 0x4f
'bbom',    # 0x50
'bbob',    # 0x51
'bbobs',    # 0x52
'bbos',    # 0x53
'bboss',    # 0x54
'bbong',    # 0x55
'bboj',    # 0x56
'bboc',    # 0x57
'bbok',    # 0x58
'bbot',    # 0x59
'bbop',    # 0x5a
'bboh',    # 0x5b
'bbwa',    # 0x5c
'bbwag',    # 0x5d
'bbwagg',    # 0x5e
'bbwags',    # 0x5f
'bbwan',    # 0x60
'bbwanj',    # 0x61
'bbwanh',    # 0x62
'bbwad',    # 0x63
'bbwal',    # 0x64
'bbwalg',    # 0x65
'bbwalm',    # 0x66
'bbwalb',    # 0x67
'bbwals',    # 0x68
'bbwalt',    # 0x69
'bbwalp',    # 0x6a
'bbwalh',    # 0x6b
'bbwam',    # 0x6c
'bbwab',    # 0x6d
'bbwabs',    # 0x6e
'bbwas',    # 0x6f
'bbwass',    # 0x70
'bbwang',    # 0x71
'bbwaj',    # 0x72
'bbwac',    # 0x73
'bbwak',    # 0x74
'bbwat',    # 0x75
'bbwap',    # 0x76
'bbwah',    # 0x77
'bbwae',    # 0x78
'bbwaeg',    # 0x79
'bbwaegg',    # 0x7a
'bbwaegs',    # 0x7b
'bbwaen',    # 0x7c
'bbwaenj',    # 0x7d
'bbwaenh',    # 0x7e
'bbwaed',    # 0x7f
'bbwael',    # 0x80
'bbwaelg',    # 0x81
'bbwaelm',    # 0x82
'bbwaelb',    # 0x83
'bbwaels',    # 0x84
'bbwaelt',    # 0x85
'bbwaelp',    # 0x86
'bbwaelh',    # 0x87
'bbwaem',    # 0x88
'bbwaeb',    # 0x89
'bbwaebs',    # 0x8a
'bbwaes',    # 0x8b
'bbwaess',    # 0x8c
'bbwaeng',    # 0x8d
'bbwaej',    # 0x8e
'bbwaec',    # 0x8f
'bbwaek',    # 0x90
'bbwaet',    # 0x91
'bbwaep',    # 0x92
'bbwaeh',    # 0x93
'bboe',    # 0x94
'bboeg',    # 0x95
'bboegg',    # 0x96
'bboegs',    # 0x97
'bboen',    # 0x98
'bboenj',    # 0x99
'bboenh',    # 0x9a
'bboed',    # 0x9b
'bboel',    # 0x9c
'bboelg',    # 0x9d
'bboelm',    # 0x9e
'bboelb',    # 0x9f
'bboels',    # 0xa0
'bboelt',    # 0xa1
'bboelp',    # 0xa2
'bboelh',    # 0xa3
'bboem',    # 0xa4
'bboeb',    # 0xa5
'bboebs',    # 0xa6
'bboes',    # 0xa7
'bboess',    # 0xa8
'bboeng',    # 0xa9
'bboej',    # 0xaa
'bboec',    # 0xab
'bboek',    # 0xac
'bboet',    # 0xad
'bboep',    # 0xae
'bboeh',    # 0xaf
'bbyo',    # 0xb0
'bbyog',    # 0xb1
'bbyogg',    # 0xb2
'bbyogs',    # 0xb3
'bbyon',    # 0xb4
'bbyonj',    # 0xb5
'bbyonh',    # 0xb6
'bbyod',    # 0xb7
'bbyol',    # 0xb8
'bbyolg',    # 0xb9
'bbyolm',    # 0xba
'bbyolb',    # 0xbb
'bbyols',    # 0xbc
'bbyolt',    # 0xbd
'bbyolp',    # 0xbe
'bbyolh',    # 0xbf
'bbyom',    # 0xc0
'bbyob',    # 0xc1
'bbyobs',    # 0xc2
'bbyos',    # 0xc3
'bbyoss',    # 0xc4
'bbyong',    # 0xc5
'bbyoj',    # 0xc6
'bbyoc',    # 0xc7
'bbyok',    # 0xc8
'bbyot',    # 0xc9
'bbyop',    # 0xca
'bbyoh',    # 0xcb
'bbu',    # 0xcc
'bbug',    # 0xcd
'bbugg',    # 0xce
'bbugs',    # 0xcf
'bbun',    # 0xd0
'bbunj',    # 0xd1
'bbunh',    # 0xd2
'bbud',    # 0xd3
'bbul',    # 0xd4
'bbulg',    # 0xd5
'bbulm',    # 0xd6
'bbulb',    # 0xd7
'bbuls',    # 0xd8
'bbult',    # 0xd9
'bbulp',    # 0xda
'bbulh',    # 0xdb
'bbum',    # 0xdc
'bbub',    # 0xdd
'bbubs',    # 0xde
'bbus',    # 0xdf
'bbuss',    # 0xe0
'bbung',    # 0xe1
'bbuj',    # 0xe2
'bbuc',    # 0xe3
'bbuk',    # 0xe4
'bbut',    # 0xe5
'bbup',    # 0xe6
'bbuh',    # 0xe7
'bbweo',    # 0xe8
'bbweog',    # 0xe9
'bbweogg',    # 0xea
'bbweogs',    # 0xeb
'bbweon',    # 0xec
'bbweonj',    # 0xed
'bbweonh',    # 0xee
'bbweod',    # 0xef
'bbweol',    # 0xf0
'bbweolg',    # 0xf1
'bbweolm',    # 0xf2
'bbweolb',    # 0xf3
'bbweols',    # 0xf4
'bbweolt',    # 0xf5
'bbweolp',    # 0xf6
'bbweolh',    # 0xf7
'bbweom',    # 0xf8
'bbweob',    # 0xf9
'bbweobs',    # 0xfa
'bbweos',    # 0xfb
'bbweoss',    # 0xfc
'bbweong',    # 0xfd
'bbweoj',    # 0xfe
'bbweoc',    # 0xff
)



# idnr.py - functions for handling South Africa ID number
# coding: utf-8
#
# Copyright (C) 2020 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ID number (South African Identity Document number).

The South African ID number is issued to individuals within South Africa. The
number consists of 13 digits and contains information about a person's date
of birth, gender and whether the person is a citizen or permanent resident.

More information:

* https://en.wikipedia.org/wiki/South_African_identity_card
* http://www.dha.gov.za/index.php/identity-documents2

>>> validate('7503305044089')
'7503305044089'
>>> validate('8503305044089')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('9125568')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> get_gender('7503305044089')
'M'
>>> get_birth_date('7503305044089')
datetime.date(1975, 3, 30)
>>> get_citizenship('7503305044089')
'citizen'
>>> format('750330 5044089')
'750330 5044 08 9'
"""

import datetime

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    return clean(number, ' ')


def get_birth_date(number):
    """Split the date parts from the number and return the date of birth.

    Since the number only uses two digits for the year, the century may be
    incorrect.
    """
    number = compact(number)
    today = datetime.date.today()
    year = int(number[0:2]) + (100 * (today.year // 100))
    month = int(number[2:4])
    day = int(number[4:6])
    if year > today.year:
        year -= 100
    try:
        return datetime.date(year, month, day)
    except ValueError:
        raise InvalidComponent()


def get_gender(number):
    """Get the gender (M/F) from the person's ID number."""
    number = compact(number)
    if number[6] in '01234':
        return 'F'
    else:
        return 'M'


def get_citizenship(number):
    """Get the citizenship status (citizen/resident) from the ID number."""
    number = compact(number)
    if number[10] == '0':
        return 'citizen'
    elif number[10] == '1':
        return 'resident'
    else:
        raise InvalidComponent()


def validate(number):
    """Check if the number is a valid South African ID number.

    This checks the length, formatting and check digit.
    """
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 13:
        raise InvalidLength()
    get_birth_date(number)
    get_citizenship(number)
    return luhn.validate(number)


def is_valid(number):
    """Check if the number is a valid South African ID number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((number[:6], number[6:10], number[10:12], number[12:]))




class FiscalizationException(Exception):
    def __init__(
        self,           # type: FiscalizationException
        error_code,     # type: int
        message,        # type: str
    ):
        # type: (...) -> None

        self.error_code = error_code
        self.message = message



# brin.py - functions for handling Brin numbers
#
# Copyright (C) 2013-2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""BRIN number (the Dutch school identification number).

The BRIN (Basisregistratie Instellingen) is a number to identify schools and
related institutions. The number consists of four alphanumeric characters,
sometimes extended with two digits to indicate the site (this complete code
is called the vestigingsnummer).

The register of these numbers can be downloaded from:
https://www.duo.nl/open_onderwijsdata/databestanden/

More information:

* https://nl.wikipedia.org/wiki/Basisregistratie_Instellingen

>>> validate('05 KO')
'05KO'
>>> validate('07NU 00')
'07NU00'
>>> validate('12KB1')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('30AJ0A')  # location code has letter
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

import re

from stdnum.exceptions import *
from stdnum.util import clean


# this regular expression is based on what was found in the online
# database: the first two digits are always numeric, followed by two
# letters and an optional two letter location identifier
_brin_re = re.compile(r'^(?P<brin>[0-9]{2}[A-Z]{2})(?P<location>[0-9]{2})?$')


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -.').upper().strip()


def validate(number):
    """Check if the number is a valid Brun number. This currently does not
    check whether the number points to a registered school."""
    number = compact(number)
    if len(number) not in (4, 6):
        raise InvalidLength()
    match = _brin_re.search(number)
    if not match:
        raise InvalidFormat()
    return number


def is_valid(number):
    """Check if the number is a valid Brun number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from logging import getLogger
from typing import Collection, Optional

from pkg_resources import (
    Distribution,
    DistributionNotFound,
    RequirementParseError,
    VersionConflict,
    get_distribution,
)

logger = getLogger(__name__)


class DependencyConflict:
    required: str = None
    found: Optional[str] = None

    def __init__(self, required, found=None):
        self.required = required
        self.found = found

    def __str__(self):
        return f'DependencyConflict: requested: "{self.required}" but found: "{self.found}"'


def get_dist_dependency_conflicts(
    dist: Distribution,
) -> Optional[DependencyConflict]:
    main_deps = dist.requires()
    instrumentation_deps = []
    for dep in dist.requires(("instruments",)):
        if dep not in main_deps:
            # we set marker to none so string representation of the dependency looks like
            #    requests ~= 1.0
            # instead of
            #    requests ~= 1.0; extra = "instruments"
            # which does not work with `get_distribution()`
            dep.marker = None
            instrumentation_deps.append(str(dep))

    return get_dependency_conflicts(instrumentation_deps)


def get_dependency_conflicts(
    deps: Collection[str],
) -> Optional[DependencyConflict]:
    for dep in deps:
        try:
            get_distribution(dep)
        except VersionConflict as exc:
            return DependencyConflict(dep, exc.dist)
        except DistributionNotFound:
            return DependencyConflict(dep)
        except RequirementParseError as exc:
            logger.warning(
                'error parsing dependency, reporting as a conflict: "%s" - %s',
                dep,
                exc,
            )
            return DependencyConflict(dep)
    return None



from xml.etree import cElementTree as eTree

from modelcommon.fiscalwrapper import FiscalizationException, FiscalizationErrorsCode
from mw_helper import format_currency
from typing import Optional


def get_error_fiscal_message(
    model,          # type: Optional[eTree.XML]
    exception,      # type: Exception
):
    # type: (...) -> Optional[str]

    if isinstance(exception, FiscalizationException):
        if exception.error_code == FiscalizationErrorsCode.FiscalMaxValueExceeded.value:
            formatted_value = format_currency(
                model=model,
                value=float(exception.message),
            )
            return "$MAXIMUM_AMOUNT_TO_FISCALIZE|{0}|{0}".format(formatted_value)

    return exception.message



# from more_itertools 9.0
def only(iterable, default=None, too_long=None):
    """If *iterable* has only one item, return it.
    If it has zero items, return *default*.
    If it has more than one item, raise the exception given by *too_long*,
    which is ``ValueError`` by default.
    >>> only([], default='missing')
    'missing'
    >>> only([1])
    1
    >>> only([1, 2])  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError: Expected exactly one item in iterable, but got 1, 2,
     and perhaps more.'
    >>> only([1, 2], too_long=TypeError)  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    TypeError
    Note that :func:`only` attempts to advance *iterable* twice to ensure there
    is only one item.  See :func:`spy` or :func:`peekable` to check
    iterable contents less destructively.
    """
    it = iter(iterable)
    first_value = next(it, default)

    try:
        second_value = next(it)
    except StopIteration:
        pass
    else:
        msg = (
            'Expected exactly one item in iterable, but got {!r}, {!r}, '
            'and perhaps more.'.format(first_value, second_value)
        )
        raise too_long or ValueError(msg)

    return first_value



# This module is provided for backwards compatibility with
# googleapis-common-protos <= 1.52.0, where this import path contained
# all of the message and gRPC definitions.

from google.longrunning.operations_proto_pb2 import *
from google.longrunning.operations_proto_pb2 import _OPERATION
from google.longrunning.operations_proto_pb2 import _OPERATION
from google.longrunning.operations_proto_pb2 import _GETOPERATIONREQUEST
from google.longrunning.operations_proto_pb2 import _LISTOPERATIONSREQUEST
from google.longrunning.operations_proto_pb2 import _LISTOPERATIONSRESPONSE
from google.longrunning.operations_proto_pb2 import _CANCELOPERATIONREQUEST
from google.longrunning.operations_proto_pb2 import _DELETEOPERATIONREQUEST
from google.longrunning.operations_proto_pb2 import _OPERATIONINFO
from google.longrunning.operations_proto_pb2 import _OPERATIONS
from google.longrunning.operations_pb2_grpc import *



import logging

from msgbus import MBEasyContext

from typing import Optional, Dict
from xml.etree import cElementTree as eTree

from tef import PrintReceipt
from utilfunctions import get_sector_name


class PrinterInteractor(object):
    def __init__(
        self,           # type: PrinterInteractor
        pos_id,         # type: str
        mb_context,     # type: MBEasyContext
    ):
        # type: (...) -> None

        self.logger = logging.getLogger("PosActions")
        self.pos_id = pos_id
        self.printer = PrintReceipt(
            mb_context=mb_context,
            logger=self.logger,
        )

    def print_tef_tender_receipt(
        self,                       # type: PrinterInteractor
        is_mobile,                  # type: bool
        order,                      # type: eTree.XML
        tender_detail,              # type: Dict
        print_customer_receipt,     # type: Optional[bool]
    ):
        # type: (...) -> bool

        sector_name = get_sector_name(order=order)
        fiscalization_type = "APOS" if is_mobile else ""
        sector_name = None if is_mobile else sector_name

        print_customer_receipt_response = self.printer.print_tef_tender_receipt(
            pos_id=self.pos_id,
            tender_detail=tender_detail,
            fiscalization_type=fiscalization_type,
            print_customer_receipt=print_customer_receipt,
            sector=sector_name,
        )

        return print_customer_receipt_response



class TPag(object):

    DEBITO = "04"
    CREDITO = "03"
    PIX = "17"
    CARTEIRA_DIGITAL = "18"
    VALE_ALIMENTACAO = "10"
    VALE_REFEICAO = "11"
    VALE_PRESENTE = "12"
    VALE_COMBUSTIVEL = "13"



# vat.py - functions for handling Irish VAT numbers
#
# Copyright (C) 2012-2016 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""VAT (Irish tax reference number).

The Irish VAT number consists of 8 or 9 digits. The number is either 7 digits
and 1 letter (optionally followed by a W for married women), 7 digits and 2
letters, or 6 digits and 2 letters or symbols (in second and last position).

>>> validate('IE 6433435F')  # pre-2013 format
'6433435F'
>>> validate('IE 6433435OA')  # 2013 format
'6433435OA'
>>> validate('6433435E')  # incorrect check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('8D79739I')  # old style number
'8D79739I'
>>> validate('8?79739J')  # incorrect old style
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> convert('1F23456T')
'0234561T'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' -').upper().strip()
    if number.startswith('IE'):
        number = number[2:]
    return number


_alphabet = 'WABCDEFGHIJKLMNOPQRSTUV'


def calc_check_digit(number):
    """Calculate the check digit. The number passed should not have the
    check digit included."""
    number = compact(number).zfill(7)
    return _alphabet[(
        sum((8 - i) * int(n) for i, n in enumerate(number[:7])) +
        9 * _alphabet.index(number[7:])) % 23]


def validate(number):
    """Check if the number provided is a valid VAT number. This checks the
    length, formatting and check digit."""
    number = compact(number)
    if not isdigits(number[:1]) or not isdigits(number[2:7]):
        raise InvalidFormat()
    if not all(x in _alphabet for x in number[7:]):
        raise InvalidFormat()
    if len(number) not in (8, 9):
        raise InvalidLength()
    if isdigits(number[:7]):
        # new system (7 digits followed by 1 or 2 letters)
        if number[7] != calc_check_digit(number[:7] + number[8:]):
            raise InvalidChecksum()
    elif number[1] in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ+*':
        # old system (second character is a letter or symbol)
        if number[7] != calc_check_digit(number[2:7] + number[0]):
            raise InvalidChecksum()
    else:
        raise InvalidFormat()
    return number


def is_valid(number):
    """Check if the number provided is a valid VAT number. This checks the
    length, formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def convert(number):
    """Convert an "old" style 8-digit VAT number where the second character
    is a letter to the new 8-digit format where only the last digit is a
    character."""
    number = compact(number)
    if len(number) == 8 and not isdigits(number[1]):
        number = '0' + number[2:7] + number[0] + number[7:]
    return number
